from coinbase.constants import API_PREFIX
from coinbase.rest.types.perpetuals_types import (
    AllocatePortfolioResponse,
    GetPerpetualsPortfolioSummaryResponse,
    GetPerpetualsPositionResponse,
    GetPortfolioBalancesResponse,
    ListPerpetualsPositionsResponse,
    OptInOutMultiAssetCollateralResponse,
)


def allocate_portfolio(
    self, portfolio_uuid: str, symbol: str, amount: str, currency: str, **kwargs
) -> AllocatePortfolioResponse:
    """
    **Allocate Portfolio**
    ________________

    [POST] https://api.coinbase.com/api/v3/brokerage/intx/allocate

    __________

    **Description:**

    Allocate more funds to an isolated position in your Perpetuals portfolio.

    __________

    **Read more on the official documentation:** `Allocate Portfolio
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_allocateportfolio>`_
    """

    endpoint = f"{API_PREFIX}/intx/allocate"

    data = {
        "portfolio_uuid": portfolio_uuid,
        "symbol": symbol,
        "amount": amount,
        "currency": currency,
    }

    return AllocatePortfolioResponse(self.post(endpoint, data=data, **kwargs))


def get_perps_portfolio_summary(
    self, portfolio_uuid: str, **kwargs
) -> GetPerpetualsPortfolioSummaryResponse:
    """
    **Get Perpetuals Portfolio Summary**
    ________________

    [GET] https://api.coinbase.com/api/v3/brokerage/intx/portfolio/{portfolio_uuid}

    __________

    **Description:**

    Get a summary of your Perpetuals portfolio.

    __________

    **Read more on the official documentation:** `Get Perpetuals Portfolio Summary
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_getintxportfoliosummary>`_
    """
    endpoint = f"{API_PREFIX}/intx/portfolio/{portfolio_uuid}"

    return GetPerpetualsPortfolioSummaryResponse(self.get(endpoint, **kwargs))


def list_perps_positions(
    self, portfolio_uuid: str, **kwargs
) -> ListPerpetualsPositionsResponse:
    """
    **List Perpetuals Positions**
    ________________

    [GET] https://api.coinbase.com/api/v3/brokerage/intx/positions/{portfolio_uuid}

    __________

    **Description:**

    Get a list of open positions in your Perpetuals portfolio.

    __________

    **Read more on the official documentation:** `List Perpetuals Positions
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_getintxpositions>`_
    """
    endpoint = f"{API_PREFIX}/intx/positions/{portfolio_uuid}"

    return ListPerpetualsPositionsResponse(self.get(endpoint, **kwargs))


def get_perps_position(
    self, portfolio_uuid: str, symbol: str, **kwargs
) -> GetPerpetualsPositionResponse:
    """
    **Get Perpetuals Position**
    ________________

    [GET] https://api.coinbase.com/api/v3/brokerage/intx/positions/{portfolio_uuid}/{symbol}

    __________

    **Description:**

    Get a specific open position in your Perpetuals portfolio

    __________

    **Read more on the official documentation:** `Get Perpetuals Positions
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_getintxposition>`_
    """
    endpoint = f"{API_PREFIX}/intx/positions/{portfolio_uuid}/{symbol}"

    return GetPerpetualsPositionResponse(self.get(endpoint, **kwargs))


def get_perps_portfolio_balances(
    self, portfolio_uuid: str, **kwargs
) -> GetPortfolioBalancesResponse:
    """
    **Get Portfolio Balances**
    ________________

    [GET] https://api.coinbase.com/api/v3/brokerage/intx/balances/{portfolio_uuid}

    __________

    **Description:**

    Get a list of asset balances on Intx for a given Portfolio

    __________

    **Read more on the official documentation:** `Get Portfolio Balances
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_getintxbalances>`_
    """
    endpoint = f"{API_PREFIX}/intx/balances/{portfolio_uuid}"

    return GetPortfolioBalancesResponse(self.get(endpoint, **kwargs))


def opt_in_or_out_multi_asset_collateral(
    self, portfolio_uuid: str, multi_asset_collateral_enabled: bool, **kwargs
) -> OptInOutMultiAssetCollateralResponse:
    """
    **Opt In or Out of Multi Asset Collateral**
    ________________

    [POST] https://api.coinbase.com/api/v3/brokerage/intx/multi_asset_collateral

    __________

    **Description:**

    Enable or Disable Multi Asset Collateral for a given Portfolio.

    __________

    **Read more on the official documentation:** `Opt In or Out of Multi Asset Collateral
    <https://docs.cloud.coinbase.com/advanced-trade-api/reference/retailbrokerageapi_intxmultiassetcollateral>`_
    """

    endpoint = f"{API_PREFIX}/intx/multi_asset_collateral"

    data = {
        "portfolio_uuid": portfolio_uuid,
        "multi_asset_collateral_enabled": multi_asset_collateral_enabled,
    }

    return OptInOutMultiAssetCollateralResponse(
        self.post(endpoint, data=data, **kwargs)
    )



import uuid
from typing import Any, Dict, List, Optional

from coinbase.constants import API_PREFIX
from coinbase.rest.types.orders_types import (
    CancelOrdersResponse,
    ClosePositionResponse,
    CreateOrderResponse,
    EditOrderPreviewResponse,
    EditOrderResponse,
    GetOrderResponse,
    ListFillsResponse,
    ListOrdersResponse,
    PreviewOrderResponse,
)


def generate_client_order_id() -> str:
    """
    :meta private:
    """
    return uuid.uuid4().hex


def create_order(
    self,
    client_order_id: str,
    product_id: str,
    side: str,
    order_configuration,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> CreateOrderResponse:
    """
    **Create Order**
    ________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders

    __________

    **Description:**

    Create an order with a specified ``product_id`` (asset-pair), ``side`` (buy/sell), etc.

    __________

    **Read more on the official documentation:** `Create Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder>`_
    """
    endpoint = f"{API_PREFIX}/orders"

    if not client_order_id:
        client_order_id = generate_client_order_id()

    data = {
        "client_order_id": client_order_id,
        "product_id": product_id,
        "side": side,
        "order_configuration": order_configuration,
        "self_trade_prevention_id": self_trade_prevention_id,
        "leverage": leverage,
        "margin_type": margin_type,
        "retail_portfolio_id": retail_portfolio_id,
    }

    return CreateOrderResponse(self.post(endpoint, data=data, **kwargs))


# Market orders
def market_order(
    self,
    client_order_id: str,
    product_id: str,
    side: str,
    quote_size: Optional[str] = None,
    base_size: Optional[str] = None,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> CreateOrderResponse:
    """
    **Market Order**
    ________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders

    __________

    **Description:**

    Place a market order to BUY or SELL the desired product at the given market price. If you wish to purchase the
    product, provide a quote_size and if you wish to sell the product, provide a base_size.

    __________

    **Read more on the official documentation:** `Create Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder>`_
    """

    market_market_ioc = {"quote_size": quote_size, "base_size": base_size}
    filtered_market_market_ioc = {
        key: value for key, value in market_market_ioc.items() if value is not None
    }

    order_configuration = {"market_market_ioc": filtered_market_market_ioc}

    return create_order(
        self,
        client_order_id,
        product_id,
        side,
        order_configuration,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def market_order_buy(
    self,
    client_order_id: str,
    product_id: str,
    quote_size: Optional[str] = None,
    base_size: Optional[str] = None,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> CreateOrderResponse:
    """
    **Create Market Order Buy**
    ____________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders

    __________

    **Description:**

    Place a market order to BUY the desired product at the given market price.

    __________

    **Read more on the official documentation:** `Create Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder>`_
    """
    return market_order(
        self,
        client_order_id,
        product_id,
        "BUY",
        quote_size=quote_size,
        base_size=base_size,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def market_order_sell(
    self,
    client_order_id: str,
    product_id: str,
    base_size: str,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> CreateOrderResponse:
    """
    **Create Market Order Sell**
    _____________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders

    __________

    **Description:**

    Place a market order to SELL the desired product at the given market price.

    __________

    **Read more on the official documentation:** `Create Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder>`_
    """
    return market_order(
        self,
        client_order_id,
        product_id,
        "SELL",
        base_size=base_size,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


# Limit IOC Orders
def limit_order_ioc(
    self,
    client_order_id: str,
    product_id: str,
    side: str,
    base_size: str,
    limit_price: str,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> CreateOrderResponse:
    """
    **Limit IOC Order**
    ________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders

    __________

    **Description:**

    Place a Limit Order with a IOC time-in-force policy. Provide the base_size (quantity of your base currency to
    spend) as well as a limit_price that indicates the maximum price at which the order should be filled.

    __________

    **Read more on the official documentation:** `Create Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder>`_
    """

    sor_limit_ioc = {"base_size": base_size, "limit_price": limit_price}

    order_configuration = {"sor_limit_ioc": sor_limit_ioc}

    return create_order(
        self,
        client_order_id,
        product_id,
        side,
        order_configuration,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def limit_order_ioc_buy(
    self,
    client_order_id: str,
    product_id: str,
    base_size: str,
    limit_price: str,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> CreateOrderResponse:
    """
    **Limit IOC Order Buy**
    ________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders

    __________

    **Description:**

    Place a Buy Limit Order with a IOC time-in-force policy. Provide the base_size (quantity of your base currency to
    spend) as well as a limit_price that indicates the maximum price at which the order should be filled.

    __________

    **Read more on the official documentation:** `Create Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder>`_
    """

    return limit_order_ioc(
        self,
        client_order_id,
        product_id,
        "BUY",
        base_size=base_size,
        limit_price=limit_price,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def limit_order_ioc_sell(
    self,
    client_order_id: str,
    product_id: str,
    base_size: str,
    limit_price: str,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> CreateOrderResponse:
    """
    **Limit IOC Order Sell**
    ________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders

    __________

    **Description:**

    Place a Sell Limit Order with a IOC time-in-force policy. Provide the base_size (quantity of your base currency to
    spend) as well as a limit_price that indicates the maximum price at which the order should be filled.

    __________

    **Read more on the official documentation:** `Create Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder>`_
    """

    return limit_order_ioc(
        self,
        client_order_id,
        product_id,
        "SELL",
        base_size=base_size,
        limit_price=limit_price,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


# Limit GTC orders
def limit_order_gtc(
    self,
    client_order_id: str,
    product_id: str,
    side: str,
    base_size: str,
    limit_price: str,
    post_only: bool = False,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> CreateOrderResponse:
    """
    **Limit Order GTC**
    ___________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders

    __________

    **Description:**

    Place a Limit Order with a GTC time-in-force policy. Provide the base_size (quantity of your base currency to
    spend) as well as a limit_price that indicates the maximum price at which the order should be filled.

    __________

    **Read more on the official documentation:** `Create Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder>`_
    """
    order_configuration = {
        "limit_limit_gtc": {
            "base_size": base_size,
            "limit_price": limit_price,
            "post_only": post_only,
        }
    }

    return create_order(
        self,
        client_order_id,
        product_id,
        side,
        order_configuration,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def limit_order_gtc_buy(
    self,
    client_order_id: str,
    product_id: str,
    base_size: str,
    limit_price: str,
    post_only: bool = False,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> CreateOrderResponse:
    """
    **Limit Order GTC Buy**
    _______________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders

    __________

    **Description:**

    Place a BUY Limit Order with a GTC time-in-force policy. Provide the base_size (quantity of your base currency to
    spend) as well as a limit_price that indicates the maximum price at which the order should be filled.

    __________

    **Read more on the official documentation:** `Create Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder>`_
    """
    return limit_order_gtc(
        self,
        client_order_id,
        product_id,
        "BUY",
        base_size=base_size,
        limit_price=limit_price,
        post_only=post_only,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def limit_order_gtc_sell(
    self,
    client_order_id: str,
    product_id: str,
    base_size: str,
    limit_price: str,
    post_only: bool = False,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> CreateOrderResponse:
    """
    **Limit Order GTC Sell**
    ________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders

    __________

    **Description:**

    Place a SELL Limit Order with a GTC time-in-force policy. Provide the base_size (quantity of your base currency to
    spend) as well as a limit_price that indicates the maximum price at which the order should be filled.

    __________

    **Read more on the official documentation:** `Create Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder>`_
    """
    return limit_order_gtc(
        self,
        client_order_id,
        product_id,
        "SELL",
        base_size=base_size,
        limit_price=limit_price,
        post_only=post_only,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


# Limit GTD orders
def limit_order_gtd(
    self,
    client_order_id: str,
    product_id: str,
    side: str,
    base_size: str,
    limit_price: str,
    end_time: str,
    post_only: bool = False,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> CreateOrderResponse:
    """
    **Limit Order GTD**
    ___________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders

    __________

    **Description:**

    Place a Limit Order with a GTD time-in-force policy. Unlike a Limit Order with a GTC time-in-force policy,
    this order type requires an end-time that indicates when this order should expire.

    __________

    **Read more on the official documentation:** `Create Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder>`_
    """
    order_configuration = {
        "limit_limit_gtd": {
            "base_size": base_size,
            "limit_price": limit_price,
            "end_time": end_time,
            "post_only": post_only,
        }
    }

    return create_order(
        self,
        client_order_id,
        product_id,
        side,
        order_configuration,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def limit_order_gtd_buy(
    self,
    client_order_id: str,
    product_id: str,
    base_size: str,
    limit_price: str,
    end_time: str,
    post_only: bool = False,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> CreateOrderResponse:
    """
    **Limit Order GTD Buy**
    _______________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders

    __________

    **Description:**

    Place a BUY Limit Order with a GTD time-in-force policy. Unlike a Limit Order with a GTC time-in-force policy,
    this order type requires an end-time that indicates when this order should expire.

    __________

    **Read more on the official documentation:** `Create Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder>`_
    """
    return limit_order_gtd(
        self,
        client_order_id,
        product_id,
        "BUY",
        base_size=base_size,
        limit_price=limit_price,
        end_time=end_time,
        post_only=post_only,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def limit_order_gtd_sell(
    self,
    client_order_id: str,
    product_id: str,
    base_size: str,
    limit_price: str,
    end_time: str,
    post_only: bool = False,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> CreateOrderResponse:
    """
    **Limit Order GTD Sell**
    ________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders

    __________

    **Description:**

    Place a SELL Limit Order with a GTD time-in-force policy. Unlike a Limit Order with a GTC time-in-force policy,
    this order type requires an end-time that indicates when this order should expire.

    __________

    **Read more on the official documentation:** `Create Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder>`_
    """
    return limit_order_gtd(
        self,
        client_order_id,
        product_id,
        "SELL",
        base_size=base_size,
        limit_price=limit_price,
        end_time=end_time,
        post_only=post_only,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


# Limit FOK Orders
def limit_order_fok(
    self,
    client_order_id: str,
    product_id: str,
    side: str,
    base_size: str,
    limit_price: str,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> CreateOrderResponse:
    """
    **Limit FOK Order**
    ________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders

    __________

    **Description:**

    Place a Limit Order with a FOK time-in-force policy. Provide the base_size (quantity of your base currency to
    spend) as well as a limit_price that indicates the maximum price at which the order should be filled.

    __________

    **Read more on the official documentation:** `Create Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder>`_
    """

    limit_limit_fok = {"base_size": base_size, "limit_price": limit_price}

    order_configuration = {"limit_limit_fok": limit_limit_fok}

    return create_order(
        self,
        client_order_id,
        product_id,
        side,
        order_configuration,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def limit_order_fok_buy(
    self,
    client_order_id: str,
    product_id: str,
    base_size: str,
    limit_price: str,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> CreateOrderResponse:
    """
    **Limit FOK Order Buy**
    ________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders

    __________

    **Description:**

    Place a Buy Limit Order with a FOK time-in-force policy. Provide the base_size (quantity of your base currency to
    spend) as well as a limit_price that indicates the maximum price at which the order should be filled.

    __________

    **Read more on the official documentation:** `Create Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder>`_
    """

    return limit_order_fok(
        self,
        client_order_id,
        product_id,
        "BUY",
        base_size=base_size,
        limit_price=limit_price,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def limit_order_fok_sell(
    self,
    client_order_id: str,
    product_id: str,
    base_size: str,
    limit_price: str,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> CreateOrderResponse:
    """
    **Limit FOK Order Sell**
    ________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders

    __________

    **Description:**

    Place a Sell Limit Order with a FOK time-in-force policy. Provide the base_size (quantity of your base currency to
    spend) as well as a limit_price that indicates the maximum price at which the order should be filled.

    __________

    **Read more on the official documentation:** `Create Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder>`_
    """

    return limit_order_fok(
        self,
        client_order_id,
        product_id,
        "SELL",
        base_size=base_size,
        limit_price=limit_price,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


# Stop-Limit GTC orders
def stop_limit_order_gtc(
    self,
    client_order_id: str,
    product_id: str,
    side: str,
    base_size: str,
    limit_price: str,
    stop_price: str,
    stop_direction: str,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> CreateOrderResponse:
    """
    **Stop-Limit Order GTC**
    ________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders

    __________

    **Description:**

    Place a Stop Limit order with a GTC time-in-force policy. Stop orders become active and wait to trigger based on
    the movement of the last trade price. The last trade price is the last price at which an order was filled.

    __________

    **Read more on the official documentation:** `Create Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder>`_
    """
    order_configuration = {
        "stop_limit_stop_limit_gtc": {
            "base_size": base_size,
            "limit_price": limit_price,
            "stop_price": stop_price,
            "stop_direction": stop_direction,
        }
    }

    return create_order(
        self,
        client_order_id,
        product_id,
        side,
        order_configuration,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def stop_limit_order_gtc_buy(
    self,
    client_order_id: str,
    product_id: str,
    base_size: str,
    limit_price: str,
    stop_price: str,
    stop_direction: str,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> CreateOrderResponse:
    """
    **Stop-Limit Order GTC Buy**
    ____________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders

    __________

    **Description:**

    Place a BUY Stop Limit order with a GTC time-in-force policy. Stop orders become active and wait to trigger based on
    the movement of the last trade price. The last trade price is the last price at which an order was filled.

    __________

    **Read more on the official documentation:** `Create Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder>`_
    """
    return stop_limit_order_gtc(
        self,
        client_order_id,
        product_id,
        "BUY",
        base_size=base_size,
        limit_price=limit_price,
        stop_price=stop_price,
        stop_direction=stop_direction,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def stop_limit_order_gtc_sell(
    self,
    client_order_id: str,
    product_id: str,
    base_size: str,
    limit_price: str,
    stop_price: str,
    stop_direction: str,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> CreateOrderResponse:
    """
    **Stop-Limit Order GTC Sell**
    _____________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders

    __________

    **Description:**

    Place a SELL Stop Limit order with a GTC time-in-force policy. Stop orders become active and wait to trigger based on
    the movement of the last trade price. The last trade price is the last price at which an order was filled.

    __________

    **Read more on the official documentation:** `Create Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder>`_
    """
    return stop_limit_order_gtc(
        self,
        client_order_id,
        product_id,
        "SELL",
        base_size=base_size,
        limit_price=limit_price,
        stop_price=stop_price,
        stop_direction=stop_direction,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


# Stop-Limit GTD orders
def stop_limit_order_gtd(
    self,
    client_order_id: str,
    product_id: str,
    side: str,
    base_size: str,
    limit_price: str,
    stop_price: str,
    end_time: str,
    stop_direction: str,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> CreateOrderResponse:
    """
    **Stop-Limit Order GTD**
    ________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders

    __________

    **Description:**

    Place a Stop Limit order with a GTD time-in-force policy. Stop orders become active and wait to trigger based on
    the movement of the last trade price. The last trade price is the last price at which an order was filled.

    __________

    **Read more on the official documentation:** `Create Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder>`_
    """
    order_configuration = {
        "stop_limit_stop_limit_gtd": {
            "base_size": base_size,
            "limit_price": limit_price,
            "stop_price": stop_price,
            "end_time": end_time,
            "stop_direction": stop_direction,
        }
    }

    return create_order(
        self,
        client_order_id,
        product_id,
        side,
        order_configuration,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def stop_limit_order_gtd_buy(
    self,
    client_order_id: str,
    product_id: str,
    base_size: str,
    limit_price: str,
    stop_price: str,
    end_time: str,
    stop_direction: str,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> CreateOrderResponse:
    """
    **Stop-Limit Order GTD Buy**
    ____________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders

    __________

    **Description:**

    Place a BUY Stop Limit order with a GTD time-in-force policy. Stop orders become active and wait to trigger based on
    the movement of the last trade price. The last trade price is the last price at which an order was filled.

    __________

    **Read more on the official documentation:** `Create Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder>`_
    """
    return stop_limit_order_gtd(
        self,
        client_order_id,
        product_id,
        "BUY",
        base_size=base_size,
        limit_price=limit_price,
        stop_price=stop_price,
        end_time=end_time,
        stop_direction=stop_direction,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def stop_limit_order_gtd_sell(
    self,
    client_order_id: str,
    product_id: str,
    base_size: str,
    limit_price: str,
    stop_price: str,
    end_time: str,
    stop_direction: str,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> CreateOrderResponse:
    """
    **Stop-Limit Order GTD Sell**
    _____________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders

    __________

    **Description:**

    Place a SELL Stop Limit order with a GTD time-in-force policy. Stop orders become active and wait to trigger based on
    the movement of the last trade price. The last trade price is the last price at which an order was filled.

    __________

    **Read more on the official documentation:** `Create Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder>`_
    """
    return stop_limit_order_gtd(
        self,
        client_order_id,
        product_id,
        "SELL",
        base_size=base_size,
        limit_price=limit_price,
        stop_price=stop_price,
        end_time=end_time,
        stop_direction=stop_direction,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


# Trigger Bracket GTC orders
def trigger_bracket_order_gtc(
    self,
    client_order_id: str,
    product_id: str,
    side: str,
    base_size: str,
    limit_price: str,
    stop_trigger_price: str,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> CreateOrderResponse:
    """
    **Trigger Bracket Order GTC**
    ________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders

    __________

    **Description:**

    Place a Trigger Bracket order with a GTC time-in-force policy. Trigger Bracket orders become active and wait to trigger based on
    the movement of the last trade price. The last trade price is the last price at which an order was filled.

    __________

    **Read more on the official documentation:** `Create Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder>`_
    """
    order_configuration = {
        "trigger_bracket_gtc": {
            "base_size": base_size,
            "limit_price": limit_price,
            "stop_trigger_price": stop_trigger_price,
        }
    }

    return create_order(
        self,
        client_order_id,
        product_id,
        side,
        order_configuration,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def trigger_bracket_order_gtc_buy(
    self,
    client_order_id: str,
    product_id: str,
    base_size: str,
    limit_price: str,
    stop_trigger_price: str,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> CreateOrderResponse:
    """
    **Trigger Bracket Order GTC Buy**
    ____________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders

    __________

    **Description:**

    Place a BUY Trigger Bracket order with a GTC time-in-force policy. Trigger Bracket orders become active and wait to trigger based on
    the movement of the last trade price. The last trade price is the last price at which an order was filled.

    __________

    **Read more on the official documentation:** `Create Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder>`_
    """
    return trigger_bracket_order_gtc(
        self,
        client_order_id,
        product_id,
        "BUY",
        base_size=base_size,
        limit_price=limit_price,
        stop_trigger_price=stop_trigger_price,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def trigger_bracket_order_gtc_sell(
    self,
    client_order_id: str,
    product_id: str,
    base_size: str,
    limit_price: str,
    stop_trigger_price: str,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> CreateOrderResponse:
    """
    **Trigger Bracket Order GTC Sell**
    _____________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders

    __________

    **Description:**

    Place a SELL Trigger Bracket order with a GTC time-in-force policy. Trigger Bracket orders become active and wait to trigger based on
    the movement of the last trade price. The last trade price is the last price at which an order was filled.

    __________

    **Read more on the official documentation:** `Create Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder>`_
    """
    return trigger_bracket_order_gtc(
        self,
        client_order_id,
        product_id,
        "SELL",
        base_size=base_size,
        limit_price=limit_price,
        stop_trigger_price=stop_trigger_price,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


# Trigger Bracket GTD orders
def trigger_bracket_order_gtd(
    self,
    client_order_id: str,
    product_id: str,
    side: str,
    base_size: str,
    limit_price: str,
    stop_trigger_price: str,
    end_time: str,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> CreateOrderResponse:
    """
    **Trigger Bracket Order GTD**
    ________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders

    __________

    **Description:**

    Place a Trigger Bracket order with a GTD time-in-force policy. Trigger Bracket orders become active and wait to trigger based on
    the movement of the last trade price. The last trade price is the last price at which an order was filled.

    __________

    **Read more on the official documentation:** `Create Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder>`_
    """
    order_configuration = {
        "trigger_bracket_gtd": {
            "base_size": base_size,
            "limit_price": limit_price,
            "stop_trigger_price": stop_trigger_price,
            "end_time": end_time,
        }
    }

    return create_order(
        self,
        client_order_id,
        product_id,
        side,
        order_configuration,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def trigger_bracket_order_gtd_buy(
    self,
    client_order_id: str,
    product_id: str,
    base_size: str,
    limit_price: str,
    stop_trigger_price: str,
    end_time: str,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> CreateOrderResponse:
    """
    **Trigger Bracket Order GTD Buy**
    ____________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders

    __________

    **Description:**

    Place a BUY Trigger Bracket order with a GTD time-in-force policy. Trigger Bracket orders become active and wait to trigger based on
    the movement of the last trade price. The last trade price is the last price at which an order was filled.

    __________

    **Read more on the official documentation:** `Create Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder>`_
    """
    return trigger_bracket_order_gtd(
        self,
        client_order_id,
        product_id,
        "BUY",
        base_size=base_size,
        limit_price=limit_price,
        stop_trigger_price=stop_trigger_price,
        end_time=end_time,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def trigger_bracket_order_gtd_sell(
    self,
    client_order_id: str,
    product_id: str,
    base_size: str,
    limit_price: str,
    stop_trigger_price: str,
    end_time: str,
    self_trade_prevention_id: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> CreateOrderResponse:
    """
    **Trigger Bracket Order GTD Sell**
    _____________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders

    __________

    **Description:**

    Place a SELL Trigger Bracket order with a GTD time-in-force policy. Trigger Bracket orders become active and wait to trigger based on
    the movement of the last trade price. The last trade price is the last price at which an order was filled.

    __________

    **Read more on the official documentation:** `Create Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_postorder>`_
    """
    return trigger_bracket_order_gtd(
        self,
        client_order_id,
        product_id,
        "SELL",
        base_size=base_size,
        limit_price=limit_price,
        stop_trigger_price=stop_trigger_price,
        end_time=end_time,
        self_trade_prevention_id=self_trade_prevention_id,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def get_order(self, order_id: str, **kwargs) -> GetOrderResponse:
    """
    **Get Order**
    _____________

    [GET] https://api.coinbase.com/api/v3/brokerage/orders/historical/{order_id}

    __________

    **Description:**

    Get a single order by order ID.

    __________

    **Read more on the official documentation:** `Get Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_gethistoricalorder>`_
    """
    endpoint = f"{API_PREFIX}/orders/historical/{order_id}"

    return GetOrderResponse(self.get(endpoint, **kwargs))


def list_orders(
    self,
    order_ids: Optional[List[str]] = None,
    product_ids: Optional[List[str]] = None,
    order_status: Optional[List[str]] = None,
    limit: Optional[int] = None,
    start_date: Optional[str] = None,
    end_date: Optional[str] = None,
    order_types: Optional[str] = None,
    order_side: Optional[str] = None,
    cursor: Optional[str] = None,
    product_type: Optional[str] = None,
    order_placement_source: Optional[str] = None,
    contract_expiry_type: Optional[str] = None,
    asset_filters: Optional[List[str]] = None,
    retail_portfolio_id: Optional[str] = None,
    time_in_forces: Optional[str] = None,
    sort_by: Optional[str] = None,
    **kwargs,
) -> ListOrdersResponse:
    """
    **List Orders**
    _______________

    [GET] https://api.coinbase.com/api/v3/brokerage/orders/historical/batch

    __________

    **Description:**

    Get a list of orders filtered by optional query parameters (``product_id``, ``order_status``, etc).

    __________

    **Read more on the official documentation:** `List Orders
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_gethistoricalorders>`_
    """
    endpoint = f"{API_PREFIX}/orders/historical/batch"
    params = {
        "order_ids": order_ids,
        "product_ids": product_ids,
        "order_status": order_status,
        "limit": limit,
        "start_date": start_date,
        "end_date": end_date,
        "order_types": order_types,
        "order_side": order_side,
        "cursor": cursor,
        "product_type": product_type,
        "order_placement_source": order_placement_source,
        "contract_expiry_type": contract_expiry_type,
        "asset_filters": asset_filters,
        "retail_portfolio_id": retail_portfolio_id,
        "time_in_forces": time_in_forces,
        "sort_by": sort_by,
    }

    return ListOrdersResponse(self.get(endpoint, params=params, **kwargs))


def get_fills(
    self,
    order_ids: Optional[List[str]] = None,
    trade_ids: Optional[List[str]] = None,
    product_ids: Optional[List[str]] = None,
    start_sequence_timestamp: Optional[str] = None,
    end_sequence_timestamp: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    limit: Optional[int] = None,
    cursor: Optional[str] = None,
    sort_by: Optional[str] = None,
    **kwargs,
) -> ListFillsResponse:
    """
    **List Fills**
    ______________

    [GET] https://api.coinbase.com/api/v3/brokerage/orders/historical/fills

    __________

    **Description:**

    Get a list of fills filtered by optional query parameters (``product_id``, ``order_id``, etc).

    __________

    **Read more on the official documentation:** `List Fills
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_getfills>`_
    """
    endpoint = f"{API_PREFIX}/orders/historical/fills"
    params = {
        "order_ids": order_ids,
        "trade_ids": trade_ids,
        "product_ids": product_ids,
        "start_sequence_timestamp": start_sequence_timestamp,
        "end_sequence_timestamp": end_sequence_timestamp,
        "retail_portfolio_id": retail_portfolio_id,
        "limit": limit,
        "cursor": cursor,
        "sort_by": sort_by,
    }

    return ListFillsResponse(self.get(endpoint, params=params, **kwargs))


def edit_order(
    self,
    order_id: str,
    size: Optional[str] = None,
    price: Optional[str] = None,
    **kwargs,
) -> EditOrderResponse:
    """
    **Edit Order**
    ______________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/edit

    __________

    **Description:**

    Edit an order with a specified new ``size``, or new ``price``. Only limit order types, with time in force type of good-till-cancelled can be edited.

    __________

    **Read more on the official documentation:** `Edit Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_editorder>`_
    """
    endpoint = f"{API_PREFIX}/orders/edit"
    data = {
        "order_id": order_id,
        "size": size,
        "price": price,
    }

    return EditOrderResponse(self.post(endpoint, data=data, **kwargs))


def preview_edit_order(
    self,
    order_id: str,
    size: Optional[str] = None,
    price: Optional[str] = None,
    **kwargs,
) -> EditOrderPreviewResponse:
    """
    **Preview Edit Order**
    ______________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/edit_preview

    __________

    **Description:**

    Simulate an edit order request with a specified new ``size``, or new ``price``, to preview the result of an edit. Only limit order types, with time in force type of good-till-cancelled can be edited.

    __________

    **Read more on the official documentation:** `Edit Order Preview
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeweditorder>`_
    """
    endpoint = f"{API_PREFIX}/orders/edit_preview"
    data = {
        "order_id": order_id,
        "size": size,
        "price": price,
    }

    return EditOrderPreviewResponse(self.post(endpoint, data=data, **kwargs))


def cancel_orders(self, order_ids: List[str], **kwargs) -> CancelOrdersResponse:
    """
    **Cancel Orders**
    _________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/batch_cancel

    __________

    **Description:**

    Initiate cancel requests for one or more orders.

    __________

    **Read more on the official documentation:** `Cancel Orders
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_cancelorders>`_
    """
    endpoint = f"{API_PREFIX}/orders/batch_cancel"
    data = {
        "order_ids": order_ids,
    }

    return CancelOrdersResponse(self.post(endpoint, data=data, **kwargs))


def preview_order(
    self,
    product_id: str,
    side: str,
    order_configuration,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> PreviewOrderResponse:
    """
    **Preview Order**
    _________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/preview

    __________

    **Description:**

    Preview the results of an order request before sending.

    __________

    **Read more on the official documentation:** `Preview Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeworder>`_
    """
    endpoint = f"{API_PREFIX}/orders/preview"

    data = {
        "product_id": product_id,
        "side": side,
        "order_configuration": order_configuration,
        "leverage": leverage,
        "margin_type": margin_type,
        "retail_portfolio_id": retail_portfolio_id,
    }

    return PreviewOrderResponse(self.post(endpoint, data=data, **kwargs))


# Preview market orders
def preview_market_order(
    self,
    product_id: str,
    side: str,
    quote_size: Optional[str] = None,
    base_size: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> PreviewOrderResponse:
    """
    **Preview Market Order**
    ________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/preview

    __________

    **Description:**

    Preview the results of a market order request before sending.

    __________

    **Read more on the official documentation:** `Preview Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeworder>`_
    """

    market_market_ioc = {"quote_size": quote_size, "base_size": base_size}
    filtered_market_market_ioc = {
        key: value for key, value in market_market_ioc.items() if value is not None
    }

    order_configuration = {"market_market_ioc": filtered_market_market_ioc}

    return preview_order(
        self,
        product_id,
        side,
        order_configuration,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def preview_market_order_buy(
    self,
    product_id: str,
    quote_size: Optional[str] = None,
    base_size: Optional[str] = None,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> PreviewOrderResponse:
    """
    **Preview Market Buy Order**
    ____________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/preview

    __________

    **Description:**

    Preview the results of a market order buy request before sending.

    __________

    **Read more on the official documentation:** `Preview Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeworder>`_
    """
    return preview_market_order(
        self,
        product_id,
        "BUY",
        quote_size=quote_size,
        base_size=base_size,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def preview_market_order_sell(
    self,
    product_id: str,
    base_size: str,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> PreviewOrderResponse:
    """
    **Preview Market Sell Order**
    _____________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/preview

    __________

    **Description:**

    Preview the results of a market order sell request before sending.

    __________

    **Read more on the official documentation:** `Preview Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeworder>`_
    """
    return preview_market_order(
        self,
        product_id,
        "SELL",
        base_size=base_size,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


# Preview Limit IOC orders
def preview_limit_order_ioc(
    self,
    product_id: str,
    side: str,
    base_size: str,
    limit_price: str,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> PreviewOrderResponse:
    """
    **Preview Limit Order IOC**
    ___________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/preview

    __________

    **Description:**

    Preview the results of a limit order IOC request before sending.

    __________

    **Read more on the official documentation:** `Preview Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeworder>`_
    """
    order_configuration = {
        "sor_limit_ioc": {"base_size": base_size, "limit_price": limit_price}
    }

    return preview_order(
        self,
        product_id,
        side,
        order_configuration,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def preview_limit_order_ioc_buy(
    self,
    product_id: str,
    base_size: str,
    limit_price: str,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> PreviewOrderResponse:
    """
    **Preview Limit Order IOC Buy**
    ___________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/preview

    __________

    **Description:**

    Preview the results of a limit order IOC buy request before sending.

    __________

    **Read more on the official documentation:** `Preview Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeworder>`_
    """
    return preview_limit_order_ioc(
        self,
        product_id,
        "BUY",
        base_size=base_size,
        limit_price=limit_price,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def preview_limit_order_ioc_sell(
    self,
    product_id: str,
    base_size: str,
    limit_price: str,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> PreviewOrderResponse:
    """
    **Preview Limit Order IOC Sell**
    ___________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/preview

    __________

    **Description:**

    Preview the results of a limit order IOC sell request before sending.

    __________

    **Read more on the official documentation:** `Preview Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeworder>`_
    """
    return preview_limit_order_ioc(
        self,
        product_id,
        "SELL",
        base_size=base_size,
        limit_price=limit_price,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


# Preview Limit GTC orders
def preview_limit_order_gtc(
    self,
    product_id: str,
    side: str,
    base_size: str,
    limit_price: str,
    post_only: bool = False,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> PreviewOrderResponse:
    """
    **Preview Limit Order GTC**
    ___________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/preview

    __________

    **Description:**

    Preview the results of a limit order GTC request before sending.

    __________

    **Read more on the official documentation:** `Preview Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeworder>`_
    """
    order_configuration = {
        "limit_limit_gtc": {
            "base_size": base_size,
            "limit_price": limit_price,
            "post_only": post_only,
        }
    }

    return preview_order(
        self,
        product_id,
        side,
        order_configuration,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def preview_limit_order_gtc_buy(
    self,
    product_id: str,
    base_size: str,
    limit_price: str,
    post_only: bool = False,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> PreviewOrderResponse:
    """
    **Preview Limit Order GTC Buy**
    _______________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/preview

    __________

    **Description:**

    Preview the results of a limit order GTC buy request before sending.

    __________

    **Read more on the official documentation:** `Preview Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeworder>`_
    """
    return preview_limit_order_gtc(
        self,
        product_id,
        "BUY",
        base_size=base_size,
        limit_price=limit_price,
        post_only=post_only,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def preview_limit_order_gtc_sell(
    self,
    product_id: str,
    base_size: str,
    limit_price: str,
    post_only: bool = False,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> PreviewOrderResponse:
    """
    **Preview Limit Order GTC Sell**
    ________________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/preview

    __________

    **Description:**

    Preview the results of a limit order GTC sell request before sending.

    __________

    **Read more on the official documentation:** `Preview Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeworder>`_
    """
    return preview_limit_order_gtc(
        self,
        product_id,
        "SELL",
        base_size=base_size,
        limit_price=limit_price,
        post_only=post_only,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


# Preview Limit GTD orders
def preview_limit_order_gtd(
    self,
    product_id: str,
    side: str,
    base_size: str,
    limit_price: str,
    end_time: str,
    post_only: bool = False,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> PreviewOrderResponse:
    """
    **Preview Limit Order GTD**
    ___________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/preview

    __________

    **Description:**

    Preview the results of a limit order GTD request before sending.

    __________

    **Read more on the official documentation:** `Preview Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeworder>`_
    """
    order_configuration = {
        "limit_limit_gtd": {
            "base_size": base_size,
            "limit_price": limit_price,
            "end_time": end_time,
            "post_only": post_only,
        }
    }

    return preview_order(
        self,
        product_id,
        side,
        order_configuration,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def preview_limit_order_gtd_buy(
    self,
    product_id: str,
    base_size: str,
    limit_price: str,
    end_time: str,
    post_only: bool = False,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> PreviewOrderResponse:
    """
    **Preview Limit Order GTD Buy**
    _______________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/preview

    __________

    **Description:**

    Preview the results of a limit order GTD buy request before sending.

    __________

    **Read more on the official documentation:** `Preview Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeworder>`_
    """
    return preview_limit_order_gtd(
        self,
        product_id,
        "BUY",
        base_size=base_size,
        limit_price=limit_price,
        end_time=end_time,
        post_only=post_only,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def preview_limit_order_gtd_sell(
    self,
    product_id: str,
    base_size: str,
    limit_price: str,
    end_time: str,
    post_only: bool = False,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> PreviewOrderResponse:
    """
    **Preview Limit Order GTD Sell**
    ________________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/preview

    __________

    **Description:**

    Preview the results of a limit order GTD sell request before sending.

    __________

    **Read more on the official documentation:** `Preview Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeworder>`_
    """
    return preview_limit_order_gtd(
        self,
        product_id,
        "SELL",
        base_size=base_size,
        limit_price=limit_price,
        end_time=end_time,
        post_only=post_only,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def preview_limit_order_fok(
    self,
    product_id: str,
    side: str,
    base_size: str,
    limit_price: str,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> PreviewOrderResponse:
    """
    **Preview Limit Order FOK**
    ___________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/preview

    __________

    **Description:**

    Preview the results of a limit order FOK request before sending.

    __________

    **Read more on the official documentation:** `Preview Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeworder>`_
    """
    order_configuration = {
        "limit_limit_fok": {"base_size": base_size, "limit_price": limit_price}
    }

    return preview_order(
        self,
        product_id,
        side,
        order_configuration,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def preview_limit_order_fok_buy(
    self,
    product_id: str,
    base_size: str,
    limit_price: str,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> PreviewOrderResponse:
    """
    **Preview Limit Order FOK Buy**
    ___________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/preview

    __________

    **Description:**

    Preview the results of a limit order FOK buy request before sending.

    __________

    **Read more on the official documentation:** `Preview Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeworder>`_
    """
    return preview_limit_order_fok(
        self,
        product_id,
        "BUY",
        base_size=base_size,
        limit_price=limit_price,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def preview_limit_order_fok_sell(
    self,
    product_id: str,
    base_size: str,
    limit_price: str,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> PreviewOrderResponse:
    """
    **Preview Limit Order FOK Sell**
    ___________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/preview

    __________

    **Description:**

    Preview the results of a limit order FOK sell request before sending.

    __________

    **Read more on the official documentation:** `Preview Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeworder>`_
    """
    return preview_limit_order_fok(
        self,
        product_id,
        "SELL",
        base_size=base_size,
        limit_price=limit_price,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


# Preview Stop-Limit GTC orders
def preview_stop_limit_order_gtc(
    self,
    product_id: str,
    side: str,
    base_size: str,
    limit_price: str,
    stop_price: str,
    stop_direction: str,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> PreviewOrderResponse:
    """
    **Preview Stop-Limit Order GTC**
    ________________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/preview

    __________

    **Description:**

    Preview the results of a stop limit GTC order request before sending.

    __________

    **Read more on the official documentation:** `Preview Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeworder>`_
    """
    order_configuration = {
        "stop_limit_stop_limit_gtc": {
            "base_size": base_size,
            "limit_price": limit_price,
            "stop_price": stop_price,
            "stop_direction": stop_direction,
        }
    }

    return preview_order(
        self,
        product_id,
        side,
        order_configuration,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def preview_stop_limit_order_gtc_buy(
    self,
    product_id: str,
    base_size: str,
    limit_price: str,
    stop_price: str,
    stop_direction: str,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> PreviewOrderResponse:
    """
    **Preview Stop-Limit Order GTC Buy**
    ____________________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/preview

    __________

    **Description:**

    Preview the results of a stop limit GTC order buy request before sending.

    __________

    **Read more on the official documentation:** `Preview Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeworder>`_
    """
    return preview_stop_limit_order_gtc(
        self,
        product_id,
        "BUY",
        base_size=base_size,
        limit_price=limit_price,
        stop_price=stop_price,
        stop_direction=stop_direction,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def preview_stop_limit_order_gtc_sell(
    self,
    product_id: str,
    base_size: str,
    limit_price: str,
    stop_price: str,
    stop_direction: str,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> PreviewOrderResponse:
    """
    **Preview Stop-Limit Order GTC Sell**
    _____________________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/preview

    __________

    **Description:**

    Preview the results of a stop limit GTC order sell request before sending.

    __________

    **Read more on the official documentation:** `Preview Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeworder>`_
    """
    return preview_stop_limit_order_gtc(
        self,
        product_id,
        "SELL",
        base_size=base_size,
        limit_price=limit_price,
        stop_price=stop_price,
        stop_direction=stop_direction,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


# Preview Stop-Limit GTD orders
def preview_stop_limit_order_gtd(
    self,
    product_id: str,
    side: str,
    base_size: str,
    limit_price: str,
    stop_price: str,
    end_time: str,
    stop_direction: str,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> PreviewOrderResponse:
    """
    **Preview Stop-Limit Order GTD**
    ________________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/preview

    __________

    **Description:**

    Preview the results of a stop limit GTD order request before sending.

    __________

    **Read more on the official documentation:** `Preview Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeworder>`_
    """
    order_configuration = {
        "stop_limit_stop_limit_gtd": {
            "base_size": base_size,
            "limit_price": limit_price,
            "stop_price": stop_price,
            "end_time": end_time,
            "stop_direction": stop_direction,
        }
    }

    return preview_order(
        self,
        product_id,
        side,
        order_configuration,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def preview_stop_limit_order_gtd_buy(
    self,
    product_id: str,
    base_size: str,
    limit_price: str,
    stop_price: str,
    end_time: str,
    stop_direction: str,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> PreviewOrderResponse:
    """
    **Preview Stop-Limit Order GTD Buy**
    ____________________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/preview

    __________

    **Description:**

    Preview the results of a stop limit GTD order buy request before sending.

    __________

    **Read more on the official documentation:** `Preview Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeworder>`_
    """
    return preview_stop_limit_order_gtd(
        self,
        product_id,
        "BUY",
        base_size=base_size,
        limit_price=limit_price,
        stop_price=stop_price,
        end_time=end_time,
        stop_direction=stop_direction,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def preview_stop_limit_order_gtd_sell(
    self,
    product_id: str,
    base_size: str,
    limit_price: str,
    stop_price: str,
    end_time: str,
    stop_direction: str,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> PreviewOrderResponse:
    """
    **Preview Stop-Limit Order GTD Sell**
    _____________________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/preview

    __________

    **Description:**

    Preview the results of a stop limit GTD order sell request before sending.

    __________

    **Read more on the official documentation:** `Preview Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeworder>`_
    """
    return preview_stop_limit_order_gtd(
        self,
        product_id,
        "SELL",
        base_size=base_size,
        limit_price=limit_price,
        stop_price=stop_price,
        end_time=end_time,
        stop_direction=stop_direction,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


# Preview Trigger Bracket GTC orders
def preview_trigger_bracket_order_gtc(
    self,
    product_id: str,
    side: str,
    base_size: str,
    limit_price: str,
    stop_trigger_price: str,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> PreviewOrderResponse:
    """
    **Preview Trigger Bracket Order GTC**
    ________________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/preview

    __________

    **Description:**

    Preview the results of a trigger bracket GTC order request before sending.

    __________

    **Read more on the official documentation:** `Preview Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeworder>`_
    """
    order_configuration = {
        "trigger_bracket_gtc": {
            "base_size": base_size,
            "limit_price": limit_price,
            "stop_trigger_price": stop_trigger_price,
        }
    }

    return preview_order(
        self,
        product_id,
        side,
        order_configuration,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def preview_trigger_bracket_order_gtc_buy(
    self,
    product_id: str,
    base_size: str,
    limit_price: str,
    stop_trigger_price: str,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> PreviewOrderResponse:
    """
    **Preview Trigger Bracket Order GTC Buy**
    ____________________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/preview

    __________

    **Description:**

    Preview the results of a trigger bracket GTC order buy request before sending.

    __________

    **Read more on the official documentation:** `Preview Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeworder>`_
    """
    return preview_trigger_bracket_order_gtc(
        self,
        product_id,
        "BUY",
        base_size=base_size,
        limit_price=limit_price,
        stop_trigger_price=stop_trigger_price,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def preview_trigger_bracket_order_gtc_sell(
    self,
    product_id: str,
    base_size: str,
    limit_price: str,
    stop_trigger_price: str,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> PreviewOrderResponse:
    """
    **Preview Trigger Bracket Order GTC Sell**
    _____________________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/preview

    __________

    **Description:**

    Preview the results of a trigger bracket GTC order sell request before sending.

    __________

    **Read more on the official documentation:** `Preview Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeworder>`_
    """
    return preview_trigger_bracket_order_gtc(
        self,
        product_id,
        "SELL",
        base_size=base_size,
        limit_price=limit_price,
        stop_trigger_price=stop_trigger_price,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


# Preview Trigger Bracket GTD orders
def preview_trigger_bracket_order_gtd(
    self,
    product_id: str,
    side: str,
    base_size: str,
    limit_price: str,
    stop_trigger_price: str,
    end_time: str,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> PreviewOrderResponse:
    """
    **Preview Trigger Bracket Order GTD**
    ________________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/preview

    __________

    **Description:**

    Preview the results of a trigger bracket GTD order request before sending.

    __________

    **Read more on the official documentation:** `Preview Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeworder>`_
    """
    order_configuration = {
        "trigger_bracket_gtd": {
            "base_size": base_size,
            "limit_price": limit_price,
            "stop_trigger_price": stop_trigger_price,
            "end_time": end_time,
        }
    }

    return preview_order(
        self,
        product_id,
        side,
        order_configuration,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def preview_trigger_bracket_order_gtd_buy(
    self,
    product_id: str,
    base_size: str,
    limit_price: str,
    stop_trigger_price: str,
    end_time: str,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> PreviewOrderResponse:
    """
    **Preview Trigger Bracket Order GTD Buy**
    ____________________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/preview

    __________

    **Description:**

    Preview the results of a trigger bracket GTD order buy request before sending.

    __________

    **Read more on the official documentation:** `Preview Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeworder>`_
    """
    return preview_trigger_bracket_order_gtd(
        self,
        product_id,
        "BUY",
        base_size=base_size,
        limit_price=limit_price,
        stop_trigger_price=stop_trigger_price,
        end_time=end_time,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def preview_trigger_bracket_order_gtd_sell(
    self,
    product_id: str,
    base_size: str,
    limit_price: str,
    stop_trigger_price: str,
    end_time: str,
    leverage: Optional[str] = None,
    margin_type: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> PreviewOrderResponse:
    """
    **Preview Trigger Bracket Order GTD Sell**
    _____________________________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/preview

    __________

    **Description:**

    Preview the results of a trigger bracket GTD order sell request before sending.

    __________

    **Read more on the official documentation:** `Preview Order
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_previeworder>`_
    """
    return preview_trigger_bracket_order_gtd(
        self,
        product_id,
        "SELL",
        base_size=base_size,
        limit_price=limit_price,
        stop_trigger_price=stop_trigger_price,
        end_time=end_time,
        leverage=leverage,
        margin_type=margin_type,
        retail_portfolio_id=retail_portfolio_id,
        **kwargs,
    )


def close_position(
    self, client_order_id: str, product_id: str, size: Optional[str] = None, **kwargs
) -> ClosePositionResponse:
    """
    **Close Position**
    _________________

    [POST] https://api.coinbase.com/api/v3/brokerage/orders/close_position

    __________

    **Description:**

    Places an order to close any open positions for a specified ``product_id``.

    __________

    **Read more on the official documentation:** `Close Position
    <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_closeposition>`_
    """
    endpoint = f"{API_PREFIX}/orders/close_position"

    if not client_order_id:
        client_order_id = generate_client_order_id()

    data = {"client_order_id": client_order_id, "product_id": product_id, "size": size}

    return ClosePositionResponse(self.post(endpoint, data=data, **kwargs))



import unittest

from requests_mock import Mocker

from coinbase.rest import RESTClient
from coinbase.rest.types.accounts_types import ListAccountsResponse

from ..constants import TEST_API_KEY, TEST_API_SECRET


class AccountsTest(unittest.TestCase):
    def test_get_accounts(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {
            "accounts": [{"uuid": "account1"}, {"name": "account2"}],
            "has_next": False,
        }

        with Mocker() as m:
            m.request(
                "GET",
                "https://api.coinbase.com/api/v3/brokerage/accounts",
                json=expected_response,
            )
            accounts = client.get_accounts(
                limit=2, cursor="abcd", retail_portfolio_id="portfolio1"
            )
            captured_request = m.request_history[0]
            self.assertEqual(
                captured_request.query,
                "limit=2&cursor=abcd&retail_portfolio_id=portfolio1",
            )
            actual_response_dict = accounts.to_dict()
            expected_response_dict = ListAccountsResponse(expected_response).to_dict()
            self.assertEqual(actual_response_dict, expected_response_dict)

    def test_get_account(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"uuid": "account1"}

        with Mocker() as m:
            m.request(
                "GET",
                "https://api.coinbase.com/api/v3/brokerage/accounts/account1",
                json=expected_response,
            )
            account = client.get_account("account1")

            captured_request = m.request_history[0]

            self.assertEqual(captured_request.query, "")
            self.assertEqual(account.__dict__, expected_response)



import asyncio
import json
import logging
import os
import ssl
import threading
import time
from multiprocessing import AuthenticationError
from typing import IO, Callable, List, Optional, Union

import backoff
import websockets

from coinbase import jwt_generator
from coinbase.api_base import APIBase, get_logger
from coinbase.constants import (
    API_ENV_KEY,
    API_SECRET_ENV_KEY,
    SUBSCRIBE_MESSAGE_TYPE,
    UNSUBSCRIBE_MESSAGE_TYPE,
    USER_AGENT,
    WS_AUTH_CHANNELS,
    WS_BASE_URL,
    WS_RETRY_BASE,
    WS_RETRY_FACTOR,
    WS_RETRY_MAX,
)

logger = get_logger("coinbase.WSClient")


class WSClientException(Exception):
    """
    **WSClientException**
    ________________________________________

    -----------------------------------------

    Exception raised for errors in the WebSocket client.
    """

    pass


class WSClientConnectionClosedException(Exception):
    """
    **WSClientConnectionClosedException**
    ________________________________________

    ----------------------------------------

    Exception raised for unexpected closure in the WebSocket client.
    """

    pass


class WSBase(APIBase):
    """
    :meta private:
    """

    def __init__(
        self,
        api_key: Optional[str] = os.getenv(API_ENV_KEY),
        api_secret: Optional[str] = os.getenv(API_SECRET_ENV_KEY),
        key_file: Optional[Union[IO, str]] = None,
        base_url=WS_BASE_URL,
        timeout: Optional[int] = None,
        max_size: Optional[int] = 10 * 1024 * 1024,
        on_message: Optional[Callable[[str], None]] = None,
        on_open: Optional[Callable[[], None]] = None,
        on_close: Optional[Callable[[], None]] = None,
        retry: Optional[bool] = True,
        verbose: Optional[bool] = False,
    ):
        super().__init__(
            api_key=api_key,
            api_secret=api_secret,
            key_file=key_file,
            base_url=base_url,
            timeout=timeout,
            verbose=verbose,
        )

        if not on_message:
            raise WSClientException("on_message callback is required.")

        if verbose:
            logger.setLevel(logging.DEBUG)

        self.max_size = max_size
        self.on_message = on_message
        self.on_open = on_open
        self.on_close = on_close

        self.websocket = None
        self.loop = None
        self.thread = None
        self._task = None

        self.retry = retry
        self._retry_max_tries = WS_RETRY_MAX
        self._retry_base = WS_RETRY_BASE
        self._retry_factor = WS_RETRY_FACTOR
        self._retry_count = 0

        self.subscriptions = {}
        self._background_exception = None
        self._retrying = False

    def open(self) -> None:
        """
        **Open Websocket**
        __________________

        ------------------------

        Open the websocket client connection.
        """
        if not self.loop or self.loop.is_closed():
            self.loop = asyncio.new_event_loop()  # Create a new event loop
            self.thread = threading.Thread(target=self.loop.run_forever)
            self.thread.daemon = True
            self.thread.start()

        self._run_coroutine_threadsafe(self.open_async())

    async def open_async(self) -> None:
        """
        **Open Websocket Async**
        ________________________

        ------------------------

        Open the websocket client connection asynchronously.
        """
        self._ensure_websocket_not_open()

        headers = self._set_headers()

        logger.debug("Connecting to %s", self.base_url)
        try:
            self.websocket = await websockets.connect(
                self.base_url,
                open_timeout=self.timeout,
                max_size=self.max_size,
                user_agent_header=USER_AGENT,
                extra_headers=headers,
                ssl=ssl.SSLContext() if self.base_url.startswith("wss://") else None,
            )
            logger.debug("Successfully connected to %s", self.base_url)

            if self.on_open:
                self.on_open()

            # Start the message handler coroutine after establishing connection
            if not self._retrying:
                self._task = asyncio.create_task(self._message_handler())

        except asyncio.TimeoutError as toe:
            self.websocket = None
            logger.error("Connection attempt timed out: %s", toe)
            raise WSClientException("Connection attempt timed out") from toe
        except (websockets.exceptions.WebSocketException, OSError) as wse:
            self.websocket = None
            logger.error("Failed to establish WebSocket connection: %s", wse)
            raise WSClientException("Failed to establish WebSocket connection") from wse

    def close(self) -> None:
        """
        **Close Websocket**
        ___________________

        ------------------------

        Close the websocket client connection.
        """
        if self.loop and not self.loop.is_closed():
            # Schedule the asynchronous close
            self._run_coroutine_threadsafe(self.close_async())
            # Stop the event loop
            self.loop.call_soon_threadsafe(self.loop.stop)
            # Wait for the thread to finish
            self.thread.join()
            # Close the event loop
            self.loop.close()
        else:
            raise WSClientException("Event loop is not running.")

    async def close_async(self) -> None:
        """
        **Close Websocket Async**
        _________________________

        ------------------------

        Close the websocket client connection asynchronously.
        """
        self._ensure_websocket_open()

        logger.debug("Closing connection to %s", self.base_url)
        try:
            await self.websocket.close()
            self.websocket = None
            self.subscriptions = {}

            logger.debug("Connection closed to %s", self.base_url)

            if self.on_close:
                self.on_close()
        except (websockets.exceptions.WebSocketException, OSError) as wse:
            logger.error("Failed to close WebSocket connection: %s", wse)
            raise WSClientException("Failed to close WebSocket connection.") from wse

    def subscribe(self, product_ids: List[str], channels: List[str]) -> None:
        """
        **Subscribe**
        _____________

        ------------------------

        Subscribe to a list of channels for a list of product ids.

        - **product_ids** - product ids to subscribe to
        - **channels** - channels to subscribe to
        """
        if self.loop and not self.loop.is_closed():
            self._run_coroutine_threadsafe(self.subscribe_async(product_ids, channels))
        else:
            raise WSClientException("Websocket Client is not open.")

    async def subscribe_async(
        self, product_ids: List[str], channels: List[str]
    ) -> None:
        """
        **Subscribe Async**
        ___________________

        ------------------------

        Async subscribe to a list of channels for a list of product ids.

        - **product_ids** - product ids to subscribe to
        - **channels** - channels to subscribe to
        """
        self._ensure_websocket_open()
        for channel in channels:
            try:
                if not self.is_authenticated and channel in WS_AUTH_CHANNELS:
                    raise AuthenticationError(
                        "Unauthenticated request to private channel."
                    )

                is_public = False if channel in WS_AUTH_CHANNELS else True
                message = self._build_subscription_message(
                    product_ids, channel, SUBSCRIBE_MESSAGE_TYPE, is_public
                )
                json_message = json.dumps(message)

                logger.debug(
                    "Subscribing to channel %s for product IDs: %s",
                    channel,
                    product_ids,
                )

                await self.websocket.send(json_message)

                logger.debug("Successfully sent subscription message.")

                # add to subscriptions map
                if channel not in self.subscriptions:
                    self.subscriptions[channel] = set()
                self.subscriptions[channel].update(product_ids)
            except websockets.exceptions.WebSocketException as wse:
                logger.error(
                    "Failed to subscribe to %s channel for product IDs %s: %s",
                    channel,
                    product_ids,
                    wse,
                )
                raise WSClientException(
                    f"Failed to subscribe to {channel} channel for product ids {product_ids}."
                ) from wse

    def unsubscribe(self, product_ids: List[str], channels: List[str]) -> None:
        """
        **Unsubscribe**
        _______________

        ------------------------

        Unsubscribe to a list of channels for a list of product ids.

        - **product_ids** - product ids to unsubscribe from
        - **channels** - channels to unsubscribe from
        """
        if self.loop and not self.loop.is_closed():
            self._run_coroutine_threadsafe(
                self.unsubscribe_async(product_ids, channels)
            )
        else:
            raise WSClientException("Websocket Client is not open.")

    async def unsubscribe_async(
        self, product_ids: List[str], channels: List[str]
    ) -> None:
        """
        **Unsubscribe Async**
        _____________________

        ------------------------

        Async unsubscribe to a list of channels for a list of product ids.

        - **product_ids** - product ids to unsubscribe from
        - **channels** - channels to unsubscribe from
        """
        self._ensure_websocket_open()
        for channel in channels:
            try:
                if not self.is_authenticated and channel in WS_AUTH_CHANNELS:
                    raise AuthenticationError(
                        "Unauthenticated request to private channel. If you wish to access private channels, you must provide your API key and secret when initializing the WSClient."
                    )
                is_public = False if channel in WS_AUTH_CHANNELS else True
                message = self._build_subscription_message(
                    product_ids, channel, UNSUBSCRIBE_MESSAGE_TYPE, is_public
                )
                json_message = json.dumps(message)

                logger.debug(
                    "Unsubscribing from channel %s for product IDs: %s",
                    channel,
                    product_ids,
                )

                await self.websocket.send(json_message)

                logger.debug("Successfully sent unsubscribe message.")

                # remove from subscriptions map
                if channel in self.subscriptions:
                    self.subscriptions[channel].difference_update(product_ids)
            except (websockets.exceptions.WebSocketException, OSError) as wse:
                logger.error(
                    "Failed to unsubscribe to %s channel for product IDs %s: %s",
                    channel,
                    product_ids,
                    wse,
                )

                raise WSClientException(
                    f"Failed to unsubscribe to {channel} channel for product ids {product_ids}."
                ) from wse

    def unsubscribe_all(self) -> None:
        """
        **Unsubscribe All**
        ________________________

        ------------------------

        Unsubscribe from all channels you are currently subscribed to.
        """
        if self.loop and not self.loop.is_closed():
            self._run_coroutine_threadsafe(self.unsubscribe_all_async())
        else:
            raise WSClientException("Websocket Client is not open.")

    async def unsubscribe_all_async(self) -> None:
        """
        **Unsubscribe All Async**
        _________________________

        ------------------------

        Async unsubscribe from all channels you are currently subscribed to.
        """
        for channel, product_ids in self.subscriptions.items():
            if product_ids:
                await self.unsubscribe_async(list(product_ids), [channel])

    def sleep_with_exception_check(self, sleep: int) -> None:
        """
        **Sleep with Exception Check**
        ______________________________

        ------------------------

        Sleep for a specified number of seconds and check for background exceptions.

        - **sleep** - number of seconds to sleep.
        """
        time.sleep(sleep)
        self.raise_background_exception()

    async def sleep_with_exception_check_async(self, sleep: int) -> None:
        """
        **Sleep with Exception Check Async**
        ____________________________________

        ------------------------

        Async sleep for a specified number of seconds and check for background exceptions.

        - **sleep** - number of seconds to sleep.
        """
        await asyncio.sleep(sleep)
        self.raise_background_exception()

    def run_forever_with_exception_check(self) -> None:
        """
        **Run Forever with Exception Check**
        ____________________________________

        ------------------------

        Runs an endless loop, checking for background exceptions every second.
        """
        while True:
            time.sleep(1)
            self.raise_background_exception()

    async def run_forever_with_exception_check_async(self) -> None:
        """
        **Run Forever with Exception Check Async**
        __________________________________________

        ------------------------

        Async runs an endless loop, checking for background exceptions every second.
        """
        while True:
            await asyncio.sleep(1)
            self.raise_background_exception()

    def raise_background_exception(self) -> None:
        """
        **Raise Background Exception**
        ______________________________

        ------------------------

        Raise any background exceptions that occurred in the message handler.
        """
        if self._background_exception:
            exception_to_raise = self._background_exception
            self._background_exception = None
            raise exception_to_raise

    def _run_coroutine_threadsafe(self, coro):
        """
        :meta private:
        """
        future = asyncio.run_coroutine_threadsafe(coro, self.loop)
        return future.result()

    def _is_websocket_open(self):
        """
        :meta private:
        """
        return self.websocket and self.websocket.open

    async def _resubscribe(self):
        """
        :meta private:
        """
        for channel, product_ids in self.subscriptions.items():
            if product_ids:
                await self.subscribe_async(list(product_ids), [channel])

    async def _retry_connection(self):
        """
        :meta private:
        """
        self._retry_count = 0

        @backoff.on_exception(
            backoff.expo,
            WSClientException,
            max_tries=self._retry_max_tries,
            base=self._retry_base,
            factor=self._retry_factor,
        )
        async def _retry_connect_and_resubscribe():
            self._retry_count += 1

            logger.debug("Retrying connection attempt %s", self._retry_count)
            if not self._is_websocket_open():
                await self.open_async()

            logger.debug("Resubscribing to channels")
            self._retry_count = 0
            await self._resubscribe()

        return await _retry_connect_and_resubscribe()

    async def _message_handler(self):
        """
        :meta private:
        """
        self.handler_open = True
        while self._is_websocket_open():
            try:
                message = await self.websocket.recv()
                if self.on_message:
                    self.on_message(message)
            except websockets.exceptions.ConnectionClosedOK as cco:
                logger.debug("Connection closed (OK): %s", cco)
                break
            except websockets.exceptions.ConnectionClosedError as cce:
                logger.error("Connection closed (ERROR): %s", cce)
                if self.retry:
                    self._retrying = True
                    try:
                        logger.debug("Retrying connection")
                        await self._retry_connection()
                        self._retrying = False
                    except WSClientException:
                        logger.error(
                            "Connection closed unexpectedly. Retry attempts failed."
                        )
                        self._background_exception = WSClientConnectionClosedException(
                            "Connection closed unexpectedly. Retry attempts failed."
                        )
                        self.subscriptions = {}
                        self._retrying = False
                        self._retry_count = 0
                        break
                else:
                    logger.error("Connection closed unexpectedly with error: %s", cce)
                    self._background_exception = WSClientConnectionClosedException(
                        f"Connection closed unexpectedly with error: {cce}"
                    )
                    self.subscriptions = {}
                    break
            except (
                websockets.exceptions.WebSocketException,
                json.JSONDecodeError,
                WSClientException,
            ) as e:
                logger.error("Exception in message handler: %s", e)
                self._background_exception = WSClientException(
                    f"Exception in message handler: {e}"
                )
                break

    def _build_subscription_message(
        self, product_ids: List[str], channel: str, message_type: str, public: bool
    ):
        """
        :meta private:
        """
        return {
            "type": message_type,
            "product_ids": product_ids,
            "channel": channel,
            **(
                {
                    "jwt": jwt_generator.build_ws_jwt(self.api_key, self.api_secret),
                }
                if self.is_authenticated
                else {}
            ),
        }

    def _ensure_websocket_not_open(self):
        """
        :meta private:
        """
        if self._is_websocket_open():
            raise WSClientException("WebSocket is already open.")

    def _ensure_websocket_open(self):
        """
        :meta private:
        """
        if not self._is_websocket_open():
            raise WSClientException("WebSocket is closed or was never opened.")

    def _set_headers(self):
        """
        :meta private:
        """
        if self._retry_count > 0:
            return {"x-cb-retry-counter": str(self._retry_count)}
        else:
            return {}



from typing import Any, Dict, List, Optional

from coinbase.rest.types.base_response import BaseResponse
from coinbase.rest.types.common_types import (
    FCMBalanceSummary,
    FCMPosition,
    FCMSweep,
    IntradayMarginSetting,
)


# Get Futures Balance Summary
class GetFuturesBalanceSummaryResponse(BaseResponse):
    def __init__(self, response: dict):
        if "balance_summary" in response:
            self.balance_summary: Optional[FCMBalanceSummary] = FCMBalanceSummary(
                **response.pop("balance_summary")
            )
        super().__init__(**response)


# List Futures Positions
class ListFuturesPositionsResponse(BaseResponse):
    def __init__(self, response: dict):
        if "positions" in response:
            self.positions: Optional[List[FCMPosition]] = [
                FCMPosition(**position) for position in response.pop("positions")
            ]
        super().__init__(**response)


# Get Futures Position
class GetFuturesPositionResponse(BaseResponse):
    def __init__(self, response: dict):
        if "position" in response:
            self.position: Optional[FCMPosition] = FCMPosition(
                **response.pop("position")
            )
        super().__init__(**response)


# Schedule Futures Sweep
class ScheduleFuturesSweepResponse(BaseResponse):
    def __init__(self, response: dict):
        if "success" in response:
            self.success: Optional[bool] = response.pop("success")
        super().__init__(**response)


# List Futures Sweeps
class ListFuturesSweepsResponse(BaseResponse):
    def __init__(self, response: dict):
        if "sweeps" in response:
            self.sweeps: List[FCMSweep] = [
                FCMSweep(**sweep) for sweep in response.pop("sweeps")
            ]
        super().__init__(**response)


# Cancel Pending Futures Sweep
class CancelPendingFuturesSweepResponse(BaseResponse):
    def __init__(self, response: dict):
        if "success" in response:
            self.success: Optional[bool] = response.pop("success")
        super().__init__(**response)


# Get Intraday Margin Setting
class GetIntradayMarginSettingResponse(BaseResponse):
    def __init__(self, response: dict):
        if "setting" in response:
            self.setting: Optional[IntradayMarginSetting] = response.pop("setting")
        super().__init__(**response)


# Get Current Margin Window
class GetCurrentMarginWindowResponse(BaseResponse):
    def __init__(self, response: dict):
        if "margin_window" in response:
            self.margin_window: Optional[Dict[str, Any]] = response.pop("margin_window")
        if "is_intraday_margin_killswitch_enabled" in response:
            self.is_intraday_margin_killswitch_enabled: Optional[bool] = response.pop(
                "is_intraday_margin_killswitch_enabled"
            )
        if "is_intraday_margin_enrollment_killswitch_enabled" in response:
            self.is_intraday_margin_enrollment_killswitch_enabled: Optional[bool] = (
                response.pop("is_intraday_margin_enrollment_killswitch_enabled")
            )
        super().__init__(**response)


# Set Intraday Margin Setting
class SetIntradayMarginSettingResponse(BaseResponse):
    def __init__(self, response: dict):
        super().__init__(**response)



from typing import List, Optional

from coinbase.rest.types.base_response import BaseResponse
from coinbase.rest.types.common_types import (
    PerpetualPortfolio,
    PortfolioBalance,
    PortfolioSummary,
    Position,
    PositionSummary,
)


# Allocate Portfolio
class AllocatePortfolioResponse(BaseResponse):
    def __init__(self, response: dict):
        super().__init__(**response)


# Get Perpetuals Portfolio Summary
class GetPerpetualsPortfolioSummaryResponse(BaseResponse):
    def __init__(self, response: dict):
        if "portfolios" in response:
            self.portfolios: Optional[List[PerpetualPortfolio]] = [
                PerpetualPortfolio(**portfolio)
                for portfolio in response.pop("portfolios")
            ]
        if "summary" in response:
            self.summary: Optional[PortfolioSummary] = PortfolioSummary(
                **response.pop("summary")
            )
        super().__init__(**response)


# List Perpetuals Positions
class ListPerpetualsPositionsResponse(BaseResponse):
    def __init__(self, response: dict):
        if "positions" in response:
            self.positions: Optional[List[Position]] = [
                Position(**pos) for pos in response.pop("positions")
            ]
        if "summary" in response:
            self.summary: Optional[PositionSummary] = PositionSummary(
                **response.pop("summary")
            )
        super().__init__(**response)


# Get Perpetuals Position
class GetPerpetualsPositionResponse(BaseResponse):
    def __init__(self, response: dict):
        if "position" in response:
            self.position: Optional[Position] = Position(**response.pop("position"))
        super().__init__(**response)


# Get Portfolio Balances
class GetPortfolioBalancesResponse(BaseResponse):
    def __init__(self, response: dict):
        if "portfolio_balances" in response:
            self.portfolio_balances: Optional[List[PortfolioBalance]] = [
                PortfolioBalance(**balance)
                for balance in response.pop("portfolio_balances")
            ]
        super().__init__(**response)


# Opt In or Out of Multi Asset Collateral
class OptInOutMultiAssetCollateralResponse(BaseResponse):
    def __init__(self, response: dict):
        if "cross_collateral_enabled" in response:
            self.cross_collateral_enabled: Optional[bool] = response.pop(
                "cross_collateral_enabled"
            )
        super().__init__(**response)



from typing import Any, Dict, List, Optional

from coinbase.rest.types.base_response import BaseResponse
from coinbase.rest.types.common_types import (
    Candle,
    HistoricalMarketTrade,
    PriceBook,
    Product,
    ProductType,
    ProductVenue,
)


# Get Best Bid/Ask
class GetBestBidAskResponse(BaseResponse):
    def __init__(self, response: dict):
        if "pricebooks" in response:
            self.pricebooks: Optional[List[PriceBook]] = [
                PriceBook(**pricebook) for pricebook in response.pop("pricebooks")
            ]
        super().__init__(**response)


# Get Product Book
class GetProductBookResponse(BaseResponse):
    def __init__(self, response: dict):
        if "pricebook" in response:
            self.pricebook: PriceBook = PriceBook(**response.pop("pricebook"))
        super().__init__(**response)


# List Products
class ListProductsResponse(BaseResponse):
    def __init__(self, response: dict):
        if "products" in response:
            self.products: Optional[List[Product]] = [
                Product(**product) for product in response.pop("products")
            ]
        if "num_products" in response:
            self.num_products: Optional[int] = response.pop("num_products")
        super().__init__(**response)


# Get Product
class GetProductResponse(BaseResponse):
    def __init__(self, response: dict):
        if "product_id" in response:
            self.product_id: str = response.pop("product_id")
        if "price" in response:
            self.price: str = response.pop("price")
        if "price_percentage_change_24h" in response:
            self.price_percentage_change_24h: str = response.pop(
                "price_percentage_change_24h"
            )
        if "volume_24h" in response:
            self.volume_24h: str = response.pop("volume_24h")
        if "volume_percentage_change_24h" in response:
            self.volume_percentage_change_24h: str = response.pop(
                "volume_percentage_change_24h"
            )
        if "base_increment" in response:
            self.base_increment: str = response.pop("base_increment")
        if "quote_increment" in response:
            self.quote_increment: str = response.pop("quote_increment")
        if "quote_min_size" in response:
            self.quote_min_size: str = response.pop("quote_min_size")
        if "quote_max_size" in response:
            self.quote_max_size: str = response.pop("quote_max_size")
        if "base_min_size" in response:
            self.base_min_size: str = response.pop("base_min_size")
        if "base_max_size" in response:
            self.base_max_size: str = response.pop("base_max_size")
        if "base_name" in response:
            self.base_name: str = response.pop("base_name")
        if "quote_name" in response:
            self.quote_name: str = response.pop("quote_name")
        if "watched" in response:
            self.watched: bool = response.pop("watched")
        if "is_disabled" in response:
            self.is_disabled: bool = response.pop("is_disabled")
        if "new" in response:
            self.new: bool = response.pop("new")
        if "status" in response:
            self.status: str = response.pop("status")
        if "cancel_only" in response:
            self.cancel_only: bool = response.pop("cancel_only")
        if "limit_only" in response:
            self.limit_only: bool = response.pop("limit_only")
        if "post_only" in response:
            self.post_only: bool = response.pop("post_only")
        if "trading_disabled" in response:
            self.trading_disabled: bool = response.pop("trading_disabled")
        if "auction_mode" in response:
            self.auction_mode: bool = response.pop("auction_mode")
        if "product_type" in response:
            self.product_type: Optional[ProductType] = response.pop("product_type")
        if "quote_currency_id" in response:
            self.quote_currency_id: Optional[str] = response.pop("quote_currency_id")
        if "base_currency_id" in response:
            self.base_currency_id: Optional[str] = response.pop("base_currency_id")
        if "fcm_trading_session_details" in response:
            self.fcm_trading_session_details: Optional[Dict[str, Any]] = response.pop(
                "fcm_trading_session_details"
            )
        if "mid_market_price" in response:
            self.mid_market_price: Optional[str] = response.pop("mid_market_price")
        if "alias" in response:
            self.alias: Optional[str] = response.pop("alias")
        if "alias_to" in response:
            self.alias_to: Optional[List[str]] = response.pop("alias_to")
        if "base_display_symbol" in response:
            self.base_display_symbol: str = response.pop("base_display_symbol")
        if "quote_display_symbol" in response:
            self.quote_display_symbol: Optional[str] = response.pop(
                "quote_display_symbol"
            )
        if "view_only" in response:
            self.view_only: Optional[bool] = response.pop("view_only")
        if "price_increment" in response:
            self.price_increment: Optional[str] = response.pop("price_increment")
        if "display_name" in response:
            self.display_name: Optional[str] = response.pop("display_name")
        if "product_venue" in response:
            self.product_venue: Optional[ProductVenue] = response.pop("product_venue")
        if "approximate_quote_24h_volume" in response:
            self.approximate_quote_24h_volume: Optional[str] = response.pop(
                "approximate_quote_24h_volume"
            )
        if "future_product_details" in response:
            self.future_product_details: Optional[Dict[str, Any]] = response.pop(
                "future_product_details"
            )
        super().__init__(**response)


# Get Product Candles
class GetProductCandlesResponse(BaseResponse):
    def __init__(self, response: dict):
        if "candles" in response:
            self.candles: Optional[List[Candle]] = [
                Candle(**candle) for candle in response.pop("candles")
            ]
        super().__init__(**response)


# Get Market Trades
class GetMarketTradesResponse(BaseResponse):
    def __init__(self, response: dict):
        if "trades" in response:
            self.trades: Optional[List[HistoricalMarketTrade]] = [
                HistoricalMarketTrade(**trade) for trade in response.pop("trades")
            ]
        if "best_bid" in response:
            self.best_bid: Optional[str] = response.pop("best_bid")
        if "best_ask" in response:
            self.best_ask: Optional[str] = response.pop("best_ask")
        super().__init__(**response)



from typing import Any

common_fields = {
    "x-ratelimit-remaining": "rate_limit_remaining",
    "x-ratelimit-reset": "rate_limit_reset",
    "x-ratelimit-limit": "rate_limit_limit",
}


class BaseResponse:
    def __init__(self, **kwargs):
        for field, formattedField in common_fields.items():
            if field in kwargs:
                setattr(self, formattedField, kwargs.pop(field))

        for key in list(kwargs.keys()):
            setattr(self, key, kwargs.pop(key))

    def __getitem__(self, key: str) -> Any:
        return self.__dict__.get(key)

    def __repr__(self):
        return str(self.__dict__)

    def to_dict(self) -> dict:
        dict_response = {}
        for key, value in self.__dict__.items():
            if isinstance(value, BaseResponse):
                dict_response[key] = value.to_dict()
            elif isinstance(value, list):
                dict_response[key] = [
                    item.to_dict() if isinstance(item, BaseResponse) else item
                    for item in value
                ]
            else:
                dict_response[key] = value
        return dict_response



import asyncio
from collections import namedtuple

import websockets


class MockWebSocketServer:
    def __init__(self, host="localhost", port=8765):
        self.start_server = None
        self.host = host
        self.port = port
        self.server = None
        self.active_websockets = set()

    async def handler(self, websocket):
        self.active_websockets.add(websocket)
        try:
            async for message in websocket:
                # Echo the message back to the client
                await websocket.send(message)
        except websockets.ConnectionClosed:
            pass
        finally:
            self.active_websockets.discard(websocket)

    def initialize_server(self):
        return websockets.serve(self.handler, self.host, self.port)

    async def start(self):
        self.start_server = self.initialize_server()
        self.server = await self.start_server

    async def stop(self):
        WebSocketTask = namedtuple("WebSocketTask", ["ws", "task"])

        tasks = [
            WebSocketTask(ws, asyncio.create_task(ws.close()))
            for ws in self.active_websockets
        ]
        await asyncio.gather(*(task.task for task in tasks))
        self.active_websockets -= {task.ws for task in tasks}

        if self.server:
            self.server.close()
            await self.server.wait_closed()  # Ensure the server is fully closed

    async def restart_with_error(self):
        await self.trigger_connection_closed_error()
        await self.stop()
        await asyncio.sleep(1)  # Short delay to ensure the port is freed up
        await self.start()

    async def trigger_connection_closed_error(self):
        WebSocketTask = namedtuple("WebSocketTask", ["ws", "task"])

        tasks = [
            WebSocketTask(
                ws, asyncio.create_task(ws.close(code=4000, reason="Abnormal closure"))
            )
            for ws in self.active_websockets
        ]
        await asyncio.gather(*(task.task for task in tasks))
        self.active_websockets -= {task.ws for task in tasks}


# Function to start the mock server
async def start_mock_server():
    server = MockWebSocketServer()
    await server.start()
    return server



from typing import Any, Dict, Optional

from coinbase.constants import API_PREFIX
from coinbase.rest.types.accounts_types import GetAccountResponse, ListAccountsResponse


def get_accounts(
    self,
    limit: Optional[int] = None,
    cursor: Optional[str] = None,
    retail_portfolio_id: Optional[str] = None,
    **kwargs,
) -> ListAccountsResponse:
    """
    **List Accounts**
    _________________
    [GET] https://api.coinbase.com/api/v3/brokerage/accounts

    __________

    **Description:**

    Get a list of authenticated accounts for the current user.

    __________

    **Read more on the official documentation:** `List Accounts <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_getaccounts>`_

    """
    endpoint = f"{API_PREFIX}/accounts"
    params = {
        "limit": limit,
        "cursor": cursor,
        "retail_portfolio_id": retail_portfolio_id,
    }

    return ListAccountsResponse(self.get(endpoint, params=params, **kwargs))


def get_account(self, account_uuid: str, **kwargs) -> GetAccountResponse:
    """

    **Get Account**
    _______________
    [GET] https://api.coinbase.com/api/v3/brokerage/accounts/{account_uuid}

    __________

    **Description:**

    Get a list of information about an account, given an account UUID.

    __________

    **Read more on the official documentation:** `Get Account <https://docs.cdp.coinbase.com/advanced-trade/reference/retailbrokerageapi_getaccount>`_
    """
    endpoint = f"{API_PREFIX}/accounts/{account_uuid}"

    return GetAccountResponse(self.get(endpoint, **kwargs))



import unittest

from requests_mock import Mocker

from coinbase.rest import RESTClient

from ..constants import TEST_API_KEY, TEST_API_SECRET


class PerpetualsTest(unittest.TestCase):
    def test_allocate_portfolio(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"key_1": "value_1", "key_2": "value_2"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/intx/allocate",
                json=expected_response,
            )
            response = client.allocate_portfolio(
                portfolio_uuid="test_uuid",
                symbol="BTC-PERP-INTX",
                amount="100",
                currency="USD",
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "portfolio_uuid": "test_uuid",
                    "symbol": "BTC-PERP-INTX",
                    "amount": "100",
                    "currency": "USD",
                },
            )
            self.assertEqual(response.__dict__, expected_response)

    def test_get_perps_portfolio_summary(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"key_1": "value_1", "key_2": "value_2"}

        with Mocker() as m:
            m.request(
                "GET",
                "https://api.coinbase.com/api/v3/brokerage/intx/portfolio/test_uuid",
                json=expected_response,
            )
            portfolios = client.get_perps_portfolio_summary(portfolio_uuid="test_uuid")

            captured_request = m.request_history[0]

            self.assertEqual(captured_request.query, "")
            self.assertEqual(portfolios.__dict__, expected_response)

    def test_list_perps_positions(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"key_1": "value_1", "key_2": "value_2"}

        with Mocker() as m:
            m.request(
                "GET",
                "https://api.coinbase.com/api/v3/brokerage/intx/positions/test_uuid",
                json=expected_response,
            )
            portfolios = client.list_perps_positions(portfolio_uuid="test_uuid")

            captured_request = m.request_history[0]

            self.assertEqual(captured_request.query, "")
            self.assertEqual(portfolios.__dict__, expected_response)

    def test_get_perps_position(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"key_1": "value_1", "key_2": "value_2"}

        with Mocker() as m:
            m.request(
                "GET",
                "https://api.coinbase.com/api/v3/brokerage/intx/positions/test_uuid/BTC-PERP-INTX",
                json=expected_response,
            )
            portfolios = client.get_perps_position(
                portfolio_uuid="test_uuid", symbol="BTC-PERP-INTX"
            )

            captured_request = m.request_history[0]

            self.assertEqual(captured_request.query, "")
            self.assertEqual(portfolios.__dict__, expected_response)

    def test_get_perps_portfolio_balances(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"key_1": "value_1", "key_2": "value_2"}

        with Mocker() as m:
            m.request(
                "GET",
                "https://api.coinbase.com/api/v3/brokerage/intx/balances/test_uuid",
                json=expected_response,
            )
            portfolios = client.get_perps_portfolio_balances(portfolio_uuid="test_uuid")

            captured_request = m.request_history[0]

            self.assertEqual(captured_request.query, "")
            self.assertEqual(portfolios.__dict__, expected_response)

    def test_opt_in_or_out_multi_asset_collateral(self):
        client = RESTClient(TEST_API_KEY, TEST_API_SECRET)

        expected_response = {"key_1": "value_1", "key_2": "value_2"}

        with Mocker() as m:
            m.request(
                "POST",
                "https://api.coinbase.com/api/v3/brokerage/intx/multi_asset_collateral",
                json=expected_response,
            )
            response = client.opt_in_or_out_multi_asset_collateral(
                portfolio_uuid="test_uuid",
                multi_asset_collateral_enabled=True,
            )

            captured_request = m.request_history[0]
            captured_json = captured_request.json()

            self.assertEqual(captured_request.query, "")
            self.assertEqual(
                captured_json,
                {
                    "portfolio_uuid": "test_uuid",
                    "multi_asset_collateral_enabled": True,
                },
            )
            self.assertEqual(response.__dict__, expected_response)



from typing import Any, Dict, Optional

from coinbase.rest.types.base_response import BaseResponse


# Get Transaction Summary
class GetTransactionSummaryResponse(BaseResponse):
    def __init__(self, response: dict):
        if "total_volume" in response:
            self.total_volume: int = response.pop("total_volume", 0.0)
        if "total_fees" in response:
            self.total_fees: int = response.pop("total_fees", 0.0)
        if "fee_tier" in response:
            self.fee_tier: Dict[str, Any] = response.pop("fee_tier")
        if "margin_rate" in response:
            self.margin_rate: Optional[Dict[str, Any]] = response.pop("margin_rate")
        if "goods_and_services_tax" in response:
            self.goods_and_services_tax: Optional[Dict[str, Any]] = response.pop(
                "goods_and_services_tax"
            )
        if "advanced_trade_only_volumes" in response:
            self.advanced_trade_only_volumes: Optional[int] = response.pop(
                "advanced_trade_only_volumes"
            )
        if "advanced_trade_only_fees" in response:
            self.advanced_trade_only_fees: Optional[int] = response.pop(
                "advanced_trade_only_fees"
            )
        if "coinbase_pro_volume" in response:  # deprecated
            self.coinbase_pro_volume: Optional[int] = response.pop(
                "coinbase_pro_volume"
            )
        if "coinbase_pro_fees" in response:  # deprecated
            self.coinbase_pro_fees: Optional[int] = response.pop("coinbase_pro_fees")
        if "total_balance" in response:
            self.total_balance: Optional[str] = response.pop("total_balance")
        if "has_promo_fee" in response:
            self.has_promo_fee: Optional[bool] = response.pop("has_promo_fee")
        super().__init__(**response)



import asyncio
import json
import unittest
from unittest.mock import AsyncMock, patch

import websockets

from coinbase.constants import SUBSCRIBE_MESSAGE_TYPE, UNSUBSCRIBE_MESSAGE_TYPE
from coinbase.websocket import WSClient, WSClientConnectionClosedException

from ..constants import TEST_API_KEY, TEST_API_SECRET
from . import mock_ws_server


class WSBaseTest(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        # set the event when on_message_mock is called
        self.message_received_event = asyncio.Event()
        self.on_message_mock = unittest.mock.Mock(
            side_effect=lambda message: self.message_received_event.set()
        )

        # set up mock websocket messages
        connection_closed_exception = websockets.ConnectionClosedOK(
            1000, "Normal closure", False
        )
        self.mock_websocket = AsyncMock()
        self.mock_websocket.recv = AsyncMock(
            side_effect=[
                "test message",
                connection_closed_exception,
                connection_closed_exception,
            ]
        )

        # initialize client
        self.ws = WSClient(
            TEST_API_KEY,
            TEST_API_SECRET,
            on_message=self.on_message_mock,
            retry=False,
        )

        # initialize public client
        self.ws_public = WSClient(
            on_message=self.on_message_mock,
            retry=False,
        )

    @patch("websockets.connect", new_callable=AsyncMock)
    def test_open_twice(self, mock_connect):
        # assert you cannot open a websocket client twice consecutively
        mock_connect.return_value = self.mock_websocket

        self.ws.open()
        self.assertIsNotNone(self.ws.websocket)

        with self.assertRaises(Exception):
            self.ws.open()

    @patch("websockets.connect", new_callable=AsyncMock)
    def test_err_after_close(self, mock_connect):
        # assert you cannot close a websocket client twice consecutively
        mock_connect.return_value = self.mock_websocket

        self.ws.open()
        self.assertIsNotNone(self.ws.websocket)

        self.ws.close()

        # assert you cannot close a websocket client twice consecutively
        with self.assertRaises(Exception):
            self.ws.close()

        # assert you cannot message a websocket client after closing
        with self.assertRaises(Exception):
            self.ws.subscribe(product_ids=["BTC-USD"], channels=["ticker"])

        with self.assertRaises(Exception):
            self.ws.unsubscribe(product_ids=["BTC-USD"], channels=["ticker"])

    def test_err_unopened(self):
        # assert you cannot close an unopened websocket client
        with self.assertRaises(Exception):
            self.ws.close()

        # assert you cannot message an unopened websocket client
        with self.assertRaises(Exception):
            self.ws.subscribe(product_ids=["BTC-USD"], channels=["ticker"])

        with self.assertRaises(Exception):
            self.ws.unsubscribe(product_ids=["BTC-USD"], channels=["ticker"])

    @patch("websockets.connect", new_callable=AsyncMock)
    def test_open_and_close(self, mock_connect):
        # assert you can open and close a websocket client
        mock_connect.return_value = self.mock_websocket

        # open
        self.ws.open()
        self.assertIsNotNone(self.ws.websocket)

        # assert on_message received
        self.on_message_mock.assert_called_once_with("test message")

        # close
        self.ws.close()
        self.mock_websocket.close.assert_awaited_once()

    @patch("websockets.connect", new_callable=AsyncMock)
    def test_open_and_close_unauthenticated(self, mock_connect):
        # assert you can open and close a websocket client
        mock_connect.return_value = self.mock_websocket

        # open
        self.ws_public.open()
        self.assertIsNotNone(self.ws_public.websocket)

        # assert on_message received
        self.on_message_mock.assert_called_once_with("test message")

        # close
        self.ws_public.close()
        self.mock_websocket.close.assert_awaited_once()

    @patch("websockets.connect", new_callable=AsyncMock)
    def test_reopen(self, mock_connect):
        # assert you can open, close, reopen and close a websocket client
        mock_connect.return_value = self.mock_websocket

        # open
        self.ws.open()
        self.assertIsNotNone(self.ws.websocket)

        # close
        self.ws.close()
        self.mock_websocket.close.assert_awaited_once()

        # reopen
        self.ws.open()
        self.assertIsNotNone(self.ws.websocket)
        self.assertTrue(self.ws.websocket.open)

        # close
        self.ws.close()
        self.assertEqual(self.mock_websocket.close.await_count, 2)

    @patch("websockets.connect", new_callable=AsyncMock)
    def test_subscribe_and_unsubscribe_channel(self, mock_connect):
        # assert you can subscribe and unsubscribe to a channel
        mock_connect.return_value = self.mock_websocket

        # open
        self.ws.open()
        self.assertIsNotNone(self.ws.websocket)

        # subscribe
        self.ws.subscribe(product_ids=["BTC-USD", "ETH-USD"], channels=["ticker"])
        self.mock_websocket.send.assert_awaited_once()

        # assert subscribe message
        subscribe = json.loads(self.mock_websocket.send.call_args_list[0][0][0])
        self.assertEqual(subscribe["type"], SUBSCRIBE_MESSAGE_TYPE)
        self.assertEqual(subscribe["product_ids"], ["BTC-USD", "ETH-USD"])
        self.assertEqual(subscribe["channel"], "ticker")

        # unsubscribe
        self.ws.unsubscribe(product_ids=["BTC-USD", "ETH-USD"], channels=["ticker"])
        self.assertEqual(self.mock_websocket.send.await_count, 2)

        # assert unsubscribe message
        unsubscribe = json.loads(self.mock_websocket.send.call_args_list[1][0][0])
        self.assertEqual(unsubscribe["type"], UNSUBSCRIBE_MESSAGE_TYPE)
        self.assertEqual(unsubscribe["product_ids"], ["BTC-USD", "ETH-USD"])
        self.assertEqual(unsubscribe["channel"], "ticker")

        # close
        self.ws.close()
        self.mock_websocket.close.assert_awaited_once()

    @patch("websockets.connect", new_callable=AsyncMock)
    def test_subscribe_and_unsubscribe_channel_unauthenticated(self, mock_connect):
        # assert you can subscribe and unsubscribe to a channel
        mock_connect.return_value = self.mock_websocket

        # open
        self.ws_public.open()
        self.assertIsNotNone(self.ws_public.websocket)

        # subscribe
        self.ws_public.subscribe(
            product_ids=["BTC-USD", "ETH-USD"], channels=["ticker"]
        )
        self.mock_websocket.send.assert_awaited_once()

        # assert subscribe message
        subscribe = json.loads(self.mock_websocket.send.call_args_list[0][0][0])
        self.assertEqual(subscribe["type"], SUBSCRIBE_MESSAGE_TYPE)
        self.assertEqual(subscribe["product_ids"], ["BTC-USD", "ETH-USD"])
        self.assertEqual(subscribe["channel"], "ticker")

        # unsubscribe
        self.ws_public.unsubscribe(
            product_ids=["BTC-USD", "ETH-USD"], channels=["ticker"]
        )
        self.assertEqual(self.mock_websocket.send.await_count, 2)

        # assert unsubscribe message
        unsubscribe = json.loads(self.mock_websocket.send.call_args_list[1][0][0])
        self.assertEqual(unsubscribe["type"], UNSUBSCRIBE_MESSAGE_TYPE)
        self.assertEqual(unsubscribe["product_ids"], ["BTC-USD", "ETH-USD"])
        self.assertEqual(unsubscribe["channel"], "ticker")

        # close
        self.ws_public.close()
        self.mock_websocket.close.assert_awaited_once()

    @patch("websockets.connect", new_callable=AsyncMock)
    def test_subscribe_and_unsubscribe_channels(self, mock_connect):
        # assert you can subscribe and unsubscribe to multiple channels
        mock_connect.return_value = self.mock_websocket

        # open
        self.ws.open()
        self.assertIsNotNone(self.ws.websocket)

        # subscribe
        self.ws.subscribe(
            product_ids=["BTC-USD", "ETH-USD"], channels=["ticker", "level2"]
        )
        self.assertEqual(self.mock_websocket.send.await_count, 2)

        # assert subscribe messages
        subscribe_1 = json.loads(self.mock_websocket.send.call_args_list[0][0][0])
        self.assertEqual(subscribe_1["type"], SUBSCRIBE_MESSAGE_TYPE)
        self.assertEqual(subscribe_1["product_ids"], ["BTC-USD", "ETH-USD"])
        self.assertEqual(subscribe_1["channel"], "ticker")

        subscribe_2 = json.loads(self.mock_websocket.send.call_args_list[1][0][0])
        self.assertEqual(subscribe_2["type"], SUBSCRIBE_MESSAGE_TYPE)
        self.assertEqual(subscribe_2["product_ids"], ["BTC-USD", "ETH-USD"])
        self.assertEqual(subscribe_2["channel"], "level2")

        # unsubscribe
        self.ws.unsubscribe(
            product_ids=["BTC-USD", "ETH-USD"], channels=["ticker", "level2"]
        )
        self.assertEqual(self.mock_websocket.send.await_count, 4)

        # assert unsubscribe messages
        unsubscribe_1 = json.loads(self.mock_websocket.send.call_args_list[2][0][0])
        self.assertEqual(unsubscribe_1["type"], UNSUBSCRIBE_MESSAGE_TYPE)
        self.assertEqual(unsubscribe_1["product_ids"], ["BTC-USD", "ETH-USD"])
        self.assertEqual(unsubscribe_1["channel"], "ticker")

        unsubscribe_2 = json.loads(self.mock_websocket.send.call_args_list[3][0][0])
        self.assertEqual(unsubscribe_2["type"], UNSUBSCRIBE_MESSAGE_TYPE)
        self.assertEqual(unsubscribe_2["product_ids"], ["BTC-USD", "ETH-USD"])
        self.assertEqual(unsubscribe_2["channel"], "level2")

        # close
        self.ws.close()
        self.mock_websocket.close.assert_awaited_once()

    @patch("websockets.connect", new_callable=AsyncMock)
    def test_subscribe_and_unsubscribe_channels_unauthenticated(self, mock_connect):
        # assert you can subscribe and unsubscribe to multiple channels
        mock_connect.return_value = self.mock_websocket

        # open
        self.ws_public.open()
        self.assertIsNotNone(self.ws_public.websocket)

        # subscribe
        self.ws_public.subscribe(
            product_ids=["BTC-USD", "ETH-USD"], channels=["ticker", "level2"]
        )
        self.assertEqual(self.mock_websocket.send.await_count, 2)

        # assert subscribe messages
        subscribe_1 = json.loads(self.mock_websocket.send.call_args_list[0][0][0])
        self.assertEqual(subscribe_1["type"], SUBSCRIBE_MESSAGE_TYPE)
        self.assertEqual(subscribe_1["product_ids"], ["BTC-USD", "ETH-USD"])
        self.assertEqual(subscribe_1["channel"], "ticker")

        subscribe_2 = json.loads(self.mock_websocket.send.call_args_list[1][0][0])
        self.assertEqual(subscribe_2["type"], SUBSCRIBE_MESSAGE_TYPE)
        self.assertEqual(subscribe_2["product_ids"], ["BTC-USD", "ETH-USD"])
        self.assertEqual(subscribe_2["channel"], "level2")

        # unsubscribe
        self.ws_public.unsubscribe(
            product_ids=["BTC-USD", "ETH-USD"], channels=["ticker", "level2"]
        )
        self.assertEqual(self.mock_websocket.send.await_count, 4)

        # assert unsubscribe messages
        unsubscribe_1 = json.loads(self.mock_websocket.send.call_args_list[2][0][0])
        self.assertEqual(unsubscribe_1["type"], UNSUBSCRIBE_MESSAGE_TYPE)
        self.assertEqual(unsubscribe_1["product_ids"], ["BTC-USD", "ETH-USD"])
        self.assertEqual(unsubscribe_1["channel"], "ticker")

        unsubscribe_2 = json.loads(self.mock_websocket.send.call_args_list[3][0][0])
        self.assertEqual(unsubscribe_2["type"], UNSUBSCRIBE_MESSAGE_TYPE)
        self.assertEqual(unsubscribe_2["product_ids"], ["BTC-USD", "ETH-USD"])
        self.assertEqual(unsubscribe_2["channel"], "level2")

        # close
        self.ws_public.close()
        self.mock_websocket.close.assert_awaited_once()

    @patch("websockets.connect", new_callable=AsyncMock)
    async def test_open_and_close_async(self, mock_connect):
        # assert you can open and close a websocket client
        mock_connect.return_value = self.mock_websocket

        # open
        await self.ws.open_async()
        self.assertIsNotNone(self.ws.websocket)

        # assert on_message received
        await self.message_received_event.wait()
        self.on_message_mock.assert_called_once_with("test message")

        # close
        await self.ws.close_async()
        self.mock_websocket.close.assert_awaited_once()

    @patch("websockets.connect", new_callable=AsyncMock)
    async def test_open_and_close_async_unauthenticated(self, mock_connect):
        # assert you can open and close a websocket client
        mock_connect.return_value = self.mock_websocket

        # open
        await self.ws_public.open_async()
        self.assertIsNotNone(self.ws_public.websocket)

        # assert on_message received
        await self.message_received_event.wait()
        self.on_message_mock.assert_called_once_with("test message")

        # close
        await self.ws_public.close_async()
        self.mock_websocket.close.assert_awaited_once()

    @patch("websockets.connect", new_callable=AsyncMock)
    async def test_reopen_async(self, mock_connect):
        # assert you can open, close, reopen and close a websocket client
        mock_connect.return_value = self.mock_websocket

        # open
        await self.ws.open_async()
        self.assertIsNotNone(self.ws.websocket)

        # close
        await self.ws.close_async()
        self.mock_websocket.close.assert_awaited_once()

        # reopen
        await self.ws.open_async()
        self.assertIsNotNone(self.ws.websocket)
        self.assertTrue(self.ws.websocket.open)

        # close
        await self.ws.close_async()
        self.assertEqual(self.mock_websocket.close.await_count, 2)

    @patch("websockets.connect", new_callable=AsyncMock)
    async def test_subscribe_and_unsubscribes_channel_async(self, mock_connect):
        # assert you can subscribe and unsubscribe to a channel
        mock_connect.return_value = self.mock_websocket

        # open
        await self.ws.open_async()
        self.assertIsNotNone(self.ws.websocket)

        # subscribe
        await self.ws.subscribe_async(
            product_ids=["BTC-USD", "ETH-USD"], channels=["ticker"]
        )
        self.mock_websocket.send.assert_awaited_once()

        # assert subscribe message
        subscribe = json.loads(self.mock_websocket.send.call_args_list[0][0][0])
        self.assertEqual(subscribe["type"], SUBSCRIBE_MESSAGE_TYPE)
        self.assertEqual(subscribe["product_ids"], ["BTC-USD", "ETH-USD"])
        self.assertEqual(subscribe["channel"], "ticker")

        # unsubscribe
        await self.ws.unsubscribe_async(
            product_ids=["BTC-USD", "ETH-USD"], channels=["ticker"]
        )
        self.assertEqual(self.mock_websocket.send.await_count, 2)

        # assert unsubscribe message
        unsubscribe = json.loads(self.mock_websocket.send.call_args_list[1][0][0])
        self.assertEqual(unsubscribe["type"], UNSUBSCRIBE_MESSAGE_TYPE)
        self.assertEqual(unsubscribe["product_ids"], ["BTC-USD", "ETH-USD"])
        self.assertEqual(unsubscribe["channel"], "ticker")

        # close
        await self.ws.close_async()
        self.mock_websocket.close.assert_awaited_once()

    @patch("websockets.connect", new_callable=AsyncMock)
    async def test_subscribe_and_unsubscribes_channel_async_unauthenticated(
        self, mock_connect
    ):
        # assert you can subscribe and unsubscribe to a channel
        mock_connect.return_value = self.mock_websocket

        # open
        await self.ws_public.open_async()
        self.assertIsNotNone(self.ws_public.websocket)

        # subscribe
        await self.ws_public.subscribe_async(
            product_ids=["BTC-USD", "ETH-USD"], channels=["ticker"]
        )
        self.mock_websocket.send.assert_awaited_once()

        # assert subscribe message
        subscribe = json.loads(self.mock_websocket.send.call_args_list[0][0][0])
        self.assertEqual(subscribe["type"], SUBSCRIBE_MESSAGE_TYPE)
        self.assertEqual(subscribe["product_ids"], ["BTC-USD", "ETH-USD"])
        self.assertEqual(subscribe["channel"], "ticker")

        # unsubscribe
        await self.ws_public.unsubscribe_async(
            product_ids=["BTC-USD", "ETH-USD"], channels=["ticker"]
        )
        self.assertEqual(self.mock_websocket.send.await_count, 2)

        # assert unsubscribe message
        unsubscribe = json.loads(self.mock_websocket.send.call_args_list[1][0][0])
        self.assertEqual(unsubscribe["type"], UNSUBSCRIBE_MESSAGE_TYPE)
        self.assertEqual(unsubscribe["product_ids"], ["BTC-USD", "ETH-USD"])
        self.assertEqual(unsubscribe["channel"], "ticker")

        # close
        await self.ws_public.close_async()
        self.mock_websocket.close.assert_awaited_once()

    def test_err_calling_private_unauthenticated(self):
        # open
        self.ws_public.open()
        self.assertIsNotNone(self.ws_public.websocket)

        # assert you cannot call a private endpoint unauthenticated
        with self.assertRaises(Exception):
            self.ws.subscribe(product_ids=["BTC-USD"], channels=["user"])


class WSDisconnectionTests(unittest.IsolatedAsyncioTestCase):
    # tests that run against a mock websocket server to simulate disconnections
    async def mock_send(self, message):
        self.messages_queue.put_nowait(message)

    async def asyncSetUp(self):
        self.messages_queue = asyncio.Queue()
        self.server = await mock_ws_server.start_mock_server()

        def on_message(msg):
            self.messages_queue.put_nowait(msg)

        self.ws = WSClient(
            TEST_API_KEY,
            TEST_API_SECRET,
            base_url="ws://localhost:8765",
            on_message=on_message,
            retry=False,
        )

    # self.ws._retry_base = 1
    # self.ws._retry_factor = 1.5
    # self.ws._retry_max = 5

    async def asyncTearDown(self):
        await self.server.stop()

    async def test_disconnect_error(self):
        # tests that client can catch a WSClientConnectionClosedException

        # open ws connection
        await self.ws.open_async()

        # trigger connection closed error
        await self.server.trigger_connection_closed_error()

        # Check for background exceptions
        with self.assertRaises(WSClientConnectionClosedException):
            await self.ws.run_forever_with_exception_check_async()

    async def test_reconnect(self):
        # tests that client can automatically reconnect after a WSClientConnectionClosedException

        self.ws.retry = True

        # Open WebSocket connection
        await self.ws.open_async()
        await self.ws.subscribe_async(
            product_ids=["BTC-USD", "ETH-USD"], channels=["ticker"]
        )

        await self.messages_queue.get()
        await self.ws.subscribe_async(product_ids=["BTC-USD"], channels=["heartbeats"])
        await self.messages_queue.get()

        # disconnect and restart the server
        await self.server.restart_with_error()

        # assert resubscribe messages
        resubscribe_1 = await self.messages_queue.get()
        resubscribe_1_json = json.loads(resubscribe_1)
        self.assertEqual(resubscribe_1_json["type"], SUBSCRIBE_MESSAGE_TYPE)
        self.assertEqual(
            sorted(resubscribe_1_json["product_ids"]), ["BTC-USD", "ETH-USD"]
        )
        self.assertEqual(resubscribe_1_json["channel"], "ticker")

        resubscribe_2 = await self.messages_queue.get()
        resubscribe_2_json = json.loads(resubscribe_2)
        self.assertEqual(resubscribe_2_json["type"], SUBSCRIBE_MESSAGE_TYPE)
        self.assertEqual(resubscribe_2_json["product_ids"], ["BTC-USD"])
        self.assertEqual(resubscribe_2_json["channel"], "heartbeats")

    async def test_reconnect_fail(self):
        # tests that client can catch WSClientConnectionClosedException after failed reconnection
        self.ws.retry = True
        self.ws._retry_max_tries = 1

        # Open WebSocket connection
        await self.ws.open_async()
        await self.ws.subscribe_async(
            product_ids=["BTC-USD", "ETH-USD"], channels=["ticker"]
        )

        await self.messages_queue.get()
        await self.ws.subscribe_async(product_ids=["BTC-USD"], channels=["heartbeats"])
        await self.messages_queue.get()

        with self.assertRaises(WSClientConnectionClosedException):
            # disconnect and restart the server
            await self.server.restart_with_error()

            # assert that client throws exception if it cannot reconnect
            await self.ws.run_forever_with_exception_check_async()
